import time
from threading import Lock, Thread

import pandas as pd

from qubx import logger
from qubx.core.basics import Instrument, MarketType
from qubx.core.lookups import lookup


class LighterInstrumentLoader:
    """
    Loads and manages Lighter market instruments.
    """

    def __init__(self, update_interval: str = "1h"):
        self._lock = Lock()
        self._update_interval = pd.Timedelta(update_interval)
        self.market_id_to_instrument: dict[int, Instrument] = {}
        self.symbol_to_market_id: dict[str, int] = {}
        self._update_instruments()
        self._thread = Thread(target=self._update_instruments_loop, daemon=True)
        self._thread.start()

    def _update_instruments_loop(self):
        while True:
            time.sleep(self._update_interval.total_seconds())
            self._update_instruments()

    def _update_instruments(self):
        try:
            instruments = lookup.find_instruments(exchange="LIGHTER", market_type=MarketType.SWAP)
        except Exception as e:
            logger.error(f"Error updating instruments: {e}")
            return

        with self._lock:
            self.instruments = instruments
            self.market_id_to_instrument = {int(instrument.exchange_symbol): instrument for instrument in instruments}
            self.symbol_to_market_id = {
                instrument.symbol: int(instrument.exchange_symbol) for instrument in instruments
            }

    def get_instrument_by_market_id(self, market_id: int) -> Instrument | None:
        with self._lock:
            return self.market_id_to_instrument.get(market_id)

    def get_instrument_by_symbol(self, symbol: str) -> Instrument | None:
        with self._lock:
            market_id = self.symbol_to_market_id.get(symbol)
            if not market_id:
                return None
            instrument = self.market_id_to_instrument.get(market_id)
            if not instrument:
                return None
            return instrument

    def get_market_id(self, symbol: str) -> int | None:
        with self._lock:
            return self.symbol_to_market_id.get(symbol)



"""
State restorer implementations.

This module provides implementations for restoring the complete strategy state
from various sources.
"""

import os
from pathlib import Path

import numpy as np
from pymongo import MongoClient

from qubx import logger
from qubx.core.basics import RestoredState
from qubx.core.utils import recognize_time
from qubx.restorers.balance import CsvBalanceRestorer, MongoDBBalanceRestorer
from qubx.restorers.interfaces import IStateRestorer
from qubx.restorers.position import CsvPositionRestorer, MongoDBPositionRestorer
from qubx.restorers.signal import CsvSignalRestorer, MongoDBSignalRestorer
from qubx.restorers.utils import find_latest_run_folder


class CsvStateRestorer(IStateRestorer):
    """
    State restorer that reads strategy state from CSV files.

    This restorer combines the functionality of CsvPositionRestorer, CsvSignalRestorer,
    and CsvBalanceRestorer to create a complete RestartState.
    """

    def __init__(
        self,
        base_dir: str | None = None,
        strategy_name: str | None = None,
        position_file_pattern: str = "*_positions.csv",
        signal_file_pattern: str = "*_signals.csv",
        targets_file_pattern: str = "*_targets.csv",
        balance_file_pattern: str = "*_balance.csv",
        lookback_days: int = 30,
    ):
        """
        Initialize the CSV state restorer.

        Args:
            base_dir: The base directory where log folders are stored.
                If None, defaults to the current working directory.
            strategy_name: Optional strategy name to filter files.
                If provided, only files matching the strategy name will be considered.
            position_file_pattern: The pattern for position CSV filenames.
            signal_file_pattern: The pattern for signal CSV filenames.
            targets_file_pattern: The pattern for target CSV filenames.
            balance_file_pattern: The pattern for balance CSV filenames.
            lookback_days: The number of days to look back for signals.
        """
        self.base_dir = Path(base_dir) if base_dir else Path(os.getcwd())
        self.strategy_name = strategy_name

        # Create individual restorers
        self.position_restorer = CsvPositionRestorer(
            base_dir=base_dir,
            file_pattern=position_file_pattern,
            strategy_name=strategy_name,
        )

        self.signal_targets_restorer = CsvSignalRestorer(
            base_dir=base_dir,
            signals_file_pattern=signal_file_pattern,
            targets_file_pattern=targets_file_pattern,
            strategy_name=strategy_name,
            lookback_days=lookback_days,
        )

        self.balance_restorer = CsvBalanceRestorer(
            base_dir=base_dir,
            file_pattern=balance_file_pattern,
            strategy_name=strategy_name,
        )

    def restore_state(self) -> RestoredState:
        """
        Restore the complete strategy state from CSV files.

        Returns:
            A RestoredState object containing positions, target positions, and balances.
        """
        # Find the latest run folder
        latest_run = find_latest_run_folder(self.base_dir)
        if not latest_run:
            logger.warning(f"No run folders found in {self.base_dir}")
            return RestoredState(
                time=np.datetime64("now"),
                positions={},
                instrument_to_signal_positions={},
                instrument_to_target_positions={},
                balances={},
            )

        logger.info(f"Restoring state from {latest_run}")

        # Restore positions, target positions, and balances
        positions = self.position_restorer.restore_positions()
        signals = self.signal_targets_restorer.restore_signals()
        targets = self.signal_targets_restorer.restore_targets()
        balances = self.balance_restorer.restore_balances()

        # Get latest position timestamp
        latest_position_timestamp = (
            max(position.last_update_time for position in positions.values()) if positions else np.datetime64("now")
        )
        if np.isnan(latest_position_timestamp):
            latest_position_timestamp = np.datetime64("now")

        # Create and return the restored state
        return RestoredState(
            time=recognize_time(latest_position_timestamp),
            positions=positions,
            instrument_to_signal_positions=signals,
            instrument_to_target_positions=targets,
            balances=balances,
        )


class MongoDBStateRestorer(IStateRestorer):
    """
    State restorer that reads strategy state from MongoDB.

    This restorer combines the functionality of MongoDBPositionRestorer,
    MongoDBSignalRestorer, and MongoDBBalanceRestorer to create a complete RestartState.
    """

    def __init__(
        self,
        strategy_name: str,
        mongo_uri: str = "mongodb://localhost:27017/",
        db_name: str = "default_logs_db",
        collection_name_prefix: str = "qubx_logs",
    ):
        self.mongo_uri = mongo_uri
        self.db_name = db_name
        self.collection_name_prefix = collection_name_prefix
        self.strategy_name = strategy_name

        self.client = MongoClient(mongo_uri)

        # Create individual restorers
        self.position_restorer = MongoDBPositionRestorer(
            strategy_name=strategy_name,
            mongo_client=self.client,
            db_name=db_name,
            collection_name=f"{collection_name_prefix}_positions",
        )

        self.signal_restorer = MongoDBSignalRestorer(
            strategy_name=strategy_name,
            mongo_client=self.client,
            db_name=db_name,
            collection_name=f"{collection_name_prefix}_signals",
        )

        self.targets_restorer = MongoDBSignalRestorer(
            strategy_name=strategy_name,
            mongo_client=self.client,
            db_name=db_name,
            collection_name=f"{collection_name_prefix}_targets",
        )

        self.balance_restorer = MongoDBBalanceRestorer(
            strategy_name=strategy_name,
            mongo_client=self.client,
            db_name=db_name,
            collection_name=f"{collection_name_prefix}_balance",
        )

    def restore_state(self) -> RestoredState:
        """
        Restore the complete strategy state from MongoDB.

        Returns:
            A RestoredState object containing positions, target positions, and balances.
        """
        mongo_collections = self.client[self.db_name].list_collection_names()
        required_suffixes = ["positions", "signals", "balance"]

        if not any(f"{self.collection_name_prefix}_{suffix}" in mongo_collections for suffix in required_suffixes):
            logger.warning(f"No logs collections found in MongodDB {self.db_name}.")
            self.client.close()
            return RestoredState(
                time=np.datetime64("now"),
                positions={},
                instrument_to_signal_positions={},
                instrument_to_target_positions={},
                balances={},
            )

        logger.info(f"Restoring state from MongoDB {self.db_name}")

        positions = self.position_restorer.restore_positions()
        signals = self.signal_restorer.restore_signals()
        targets = self.targets_restorer.restore_targets()
        balances = self.balance_restorer.restore_balances()

        latest_position_timestamp = (
            max(position.last_update_time for position in positions.values()) if positions else np.datetime64("now")
        )
        if np.isnan(latest_position_timestamp):
            latest_position_timestamp = np.datetime64("now")

        self.client.close()
        return RestoredState(
            time=recognize_time(latest_position_timestamp),
            positions=positions,
            instrument_to_signal_positions=signals,
            instrument_to_target_positions=targets,
            balances=balances,
        )



"""
Prometheus Metric Emitter.

This module provides an implementation of IMetricEmitter that exports metrics to Prometheus.
"""

from typing import Any, Dict, List, Literal, Optional

from prometheus_client import REGISTRY, Counter, Gauge, Summary, push_to_gateway

from qubx import logger
from qubx.core.basics import Signal, dt_64
from qubx.core.interfaces import IAccountViewer, IStrategyContext
from qubx.emitters.base import BaseMetricEmitter

# Define metric types
MetricType = Literal["gauge", "counter", "summary"]


class PrometheusMetricEmitter(BaseMetricEmitter):
    """
    Emits metrics to Prometheus using the Prometheus client library.

    This emitter can push metrics to a Prometheus Pushgateway or expose them
    via an HTTP endpoint for scraping.
    """

    def __init__(
        self,
        strategy_name: str,
        pushgateway_url: Optional[str] = None,
        expose_http: bool = False,
        http_port: int = 8000,
        namespace: str = "qubx",
        stats_to_emit: Optional[List[str]] = None,
        stats_interval: str = "1m",
        tags: Optional[Dict[str, str]] = None,
    ):
        """
        Initialize the Prometheus Metric Emitter.

        Args:
            strategy_name: Name of the strategy (used in metric names)
            pushgateway_url: URL of the Prometheus Pushgateway (optional)
            expose_http: Whether to expose metrics via HTTP endpoint
            http_port: Port to expose HTTP metrics on
            namespace: Namespace to prefix all metric names with
            stats_to_emit: Optional list of specific stats to emit
            stats_interval: Interval for emitting strategy stats (default: "1m")
            tags: Dictionary of default tags/labels to include with all metrics
        """
        # Initialize default tags with strategy name
        default_tags = tags or {}
        default_tags["strategy"] = strategy_name

        super().__init__(stats_to_emit, stats_interval, default_tags)

        self._strategy_name = strategy_name
        self._pushgateway_url = pushgateway_url
        self._namespace = namespace
        self._registry = REGISTRY

        # Cache for created metrics to avoid recreating them
        self._gauges: Dict[str, Gauge] = {}
        self._counters: Dict[str, Counter] = {}
        self._summaries: Dict[str, Summary] = {}

        if expose_http:
            from prometheus_client import start_http_server

            start_http_server(http_port)
            logger.info(f"[PrometheusMetricEmitter] Started HTTP server on port {http_port}")

        logger.info(
            f"[PrometheusMetricEmitter] Initialized for strategy '{strategy_name}'"
            f"{' with Pushgateway at ' + pushgateway_url if pushgateway_url else ''}"
        )

    def _get_or_create_gauge(self, name: str, labels: Optional[Dict[str, str]] = None) -> Gauge:
        """
        Get or create a Prometheus gauge.

        Args:
            name: Name of the gauge
            labels: Dictionary of labels for the gauge

        Returns:
            Gauge: The Prometheus gauge
        """
        label_dict = labels or {}
        key = f"{name}_{sorted(label_dict.keys())}"
        if key not in self._gauges:
            self._gauges[key] = Gauge(
                f"{self._namespace}_{name}",
                f"{name.replace('_', ' ')} metric",
                list(label_dict.keys()),
                registry=self._registry,
            )
        return self._gauges[key]

    def _get_or_create_counter(self, name: str, labels: Optional[Dict[str, str]] = None) -> Counter:
        """
        Get or create a Prometheus counter.

        Args:
            name: Name of the counter
            labels: Dictionary of labels for the counter

        Returns:
            Counter: The Prometheus counter
        """
        label_dict = labels or {}
        key = f"{name}_{sorted(label_dict.keys())}"
        if key not in self._counters:
            self._counters[key] = Counter(
                f"{self._namespace}_{name}",
                f"{name.replace('_', ' ')} metric",
                list(label_dict.keys()),
                registry=self._registry,
            )
        return self._counters[key]

    def _get_or_create_summary(self, name: str, labels: Optional[Dict[str, str]] = None) -> Summary:
        """
        Get or create a Prometheus summary.

        Args:
            name: Name of the summary
            labels: Dictionary of labels for the summary

        Returns:
            Summary: The Prometheus summary
        """
        label_dict = labels or {}
        key = f"{name}_{sorted(label_dict.keys())}"
        if key not in self._summaries:
            self._summaries[key] = Summary(
                f"{self._namespace}_{name}",
                f"{name.replace('_', ' ')} metric",
                list(label_dict.keys()),
                registry=self._registry,
            )
        return self._summaries[key]

    def _emit_impl(self, name: str, value: float, tags: Dict[str, str], timestamp: dt_64 | None = None) -> None:
        """
        Implementation of emit for Prometheus.

        Args:
            name: Name of the metric
            value: Value of the metric
            tags: Dictionary of tags/labels for the metric (already merged with default tags)
            timestamp: Optional timestamp for the metric (ignored in Prometheus)
        """
        try:
            # Extract metric type from tags if present, default to gauge
            metric_type = tags.pop("metric_type", "gauge") if "metric_type" in tags else "gauge"

            if metric_type == "counter":
                counter = self._get_or_create_counter(name, tags)
                counter.labels(**tags).inc(value)
            elif metric_type == "summary":
                summary = self._get_or_create_summary(name, tags)
                summary.labels(**tags).observe(value)
            else:  # Default to gauge
                gauge = self._get_or_create_gauge(name, tags)
                gauge.labels(**tags).set(value)

            # Push to gateway if configured
            if self._pushgateway_url:
                push_to_gateway(
                    self._pushgateway_url, job=f"{self._namespace}_{self._strategy_name}", registry=self._registry
                )
        except Exception as e:
            logger.error(f"[PrometheusMetricEmitter] Failed to emit metric {name}: {e}")

    def emit(
        self,
        name: str,
        value: float,
        tags: dict[str, Any] | None = None,
        timestamp: dt_64 | None = None,
        metric_type: MetricType = "gauge",
    ) -> None:
        """
        Emit a metric to Prometheus.

        Args:
            name: Name of the metric
            value: Value of the metric
            tags: Dictionary of tags/labels for the metric
            timestamp: Optional timestamp for the metric (ignored in Prometheus)
            metric_type: Type of metric (gauge, counter, summary)
        """
        # Add metric type to tags
        merged_tags = self._merge_tags(tags)
        merged_tags["metric_type"] = metric_type

        # Call the implementation
        self._emit_impl(name, value, merged_tags, timestamp)

    def emit_strategy_stats(self, context: IStrategyContext) -> None:
        """
        Emit standard strategy statistics to Prometheus.

        This method overrides the base implementation to add Prometheus-specific
        functionality like pushing to the Pushgateway.

        Args:
            context: The strategy context to get statistics from
        """
        # Call the parent implementation to emit the standard stats
        super().emit_strategy_stats(context)

        # Push all metrics to Pushgateway in one go if configured
        if self._pushgateway_url:
            try:
                push_to_gateway(
                    self._pushgateway_url, job=f"{self._namespace}_{self._strategy_name}", registry=self._registry
                )
            except Exception as e:
                logger.error(f"[PrometheusMetricEmitter] Failed to push metrics to gateway: {e}")

    def emit_signals(self, time: dt_64, signals: list[Signal], account: IAccountViewer) -> None:
        """
        Emit signals as Prometheus metrics.

        Args:
            time: Timestamp when the signals were generated
            signals: List of signals to emit
            account: Account viewer to get account information
        """
        if not signals:
            return

        try:
            for signal in signals:
                # Create labels for the signal
                labels = {
                    "symbol": signal.instrument.symbol,
                    "exchange": signal.instrument.exchange,
                    "group": signal.group if signal.group else "default",
                    "is_service": str(signal.is_service).lower(),
                }

                # Emit the signal value as a gauge
                gauge = self._get_or_create_gauge("signal_value", labels)
                gauge.labels(**labels).set(signal.signal)

                # Emit price-related metrics if available
                if signal.price is not None:
                    price_gauge = self._get_or_create_gauge("signal_price", labels)
                    price_gauge.labels(**labels).set(signal.price)

                if signal.stop is not None:
                    stop_gauge = self._get_or_create_gauge("signal_stop", labels)
                    stop_gauge.labels(**labels).set(signal.stop)

                if signal.take is not None:
                    take_gauge = self._get_or_create_gauge("signal_take", labels)
                    take_gauge.labels(**labels).set(signal.take)

                if signal.reference_price is not None:
                    ref_price_gauge = self._get_or_create_gauge("signal_reference_price", labels)
                    ref_price_gauge.labels(**labels).set(signal.reference_price)

            # Push to gateway if configured
            if self._pushgateway_url:
                try:
                    push_to_gateway(
                        self._pushgateway_url, job=f"{self._namespace}_{self._strategy_name}", registry=self._registry
                    )
                except Exception as e:
                    logger.error(f"[PrometheusMetricEmitter] Failed to push signal metrics to gateway: {e}")

        except Exception as e:
            logger.error(f"[PrometheusMetricEmitter] Failed to emit signals: {e}")



import asyncio
import contextlib
import json
import random
import traceback
from collections import defaultdict, deque
from dataclasses import dataclass
from enum import Enum
from time import monotonic
from typing import Any, Awaitable, Callable, Optional, cast

import orjson
import websockets
from websockets.asyncio.client import ClientConnection
from websockets.exceptions import (
    ConnectionClosed,
    ConnectionClosedError,
    ConnectionClosedOK,
)

from qubx import logger


class ConnectionState(Enum):
    DISCONNECTED = "disconnected"
    CONNECTING = "connecting"
    CONNECTED = "connected"
    RECONNECTING = "reconnecting"
    CLOSING = "closing"
    CLOSED = "closed"


@dataclass
class ReconnectionConfig:
    enabled: bool = True
    max_retries: int = 10
    initial_delay: float = 1.0
    max_delay: float = 60.0
    exponential_base: float = 2.0
    jitter: float = 0.5  # add up to +/- 0.5s jitter


@dataclass
class ChannelSubscription:
    channel: str
    handler: Callable[[dict], Awaitable[None]]
    params: dict[str, Any]


class MessageRateTracker:
    """
    Time-bucketed message rate tracker.
    """

    def __init__(self, max_window_seconds: int = 900, bucket_size: float = 1.0):
        self.bucket_size = bucket_size
        self.max_window = max_window_seconds
        self._buckets: dict[int, int] = defaultdict(int)  # {bucket_id: count}
        self._bucket_times: deque[int] = deque()  # Ordered bucket IDs for cleanup
        self._total_messages = 0

    def record_message(self) -> None:
        bucket_id = int(monotonic() / self.bucket_size)
        if self._buckets[bucket_id] == 0:
            self._bucket_times.append(bucket_id)
        self._buckets[bucket_id] += 1
        self._total_messages += 1
        self._cleanup()

    def get_count(self, window_seconds: float) -> int:
        current_bucket = int(monotonic() / self.bucket_size)
        oldest_bucket = current_bucket - int(window_seconds / self.bucket_size)
        return sum(count for bucket_id, count in self._buckets.items() if bucket_id > oldest_bucket)

    def get_rate(self, window_seconds: float) -> float:
        count = self.get_count(window_seconds)
        return count / window_seconds if window_seconds > 0 else 0.0

    @property
    def total_messages(self) -> int:
        return self._total_messages

    @property
    def bucket_count(self) -> int:
        return len(self._buckets)

    def _cleanup(self) -> None:
        cutoff = int(monotonic() / self.bucket_size) - int(self.max_window / self.bucket_size)
        while self._bucket_times and self._bucket_times[0] < cutoff:
            old_bucket = self._bucket_times.popleft()
            del self._buckets[old_bucket]


class BaseWebSocketManager:
    """
    Clean base WS manager with:
      - decoupled reader (recv-only) and worker(s) (parse+dispatch)
      - bounded inbox with drop-oldest on overflow (keeps latest)
      - optional app-level heartbeat hook
      - automatic reconnect + resubscribe
    """

    def __init__(
        self,
        url: str,
        *,
        reconnection: Optional[ReconnectionConfig] = None,
        # Disable protocol pings by default; subclasses may override or add app pings
        ping_interval: Optional[float] = None,
        ping_timeout: Optional[float] = None,
        max_size: Optional[int] = 16 * 1024 * 1024,
        # Let websockets keep reading; we enforce our own bounded inbox
        max_queue: Optional[int] = None,
        compression: Optional[str] = None,
        # Internal queue & workers
        inbox_size: int = 5000,
        workers: int = 1,
        # Optional app-level heartbeat
        app_ping_interval: Optional[float] = None,  # seconds; if None, no app pings
        no_rx_reconnect_after: Optional[float] = 60.0,  # if no inbound traffic for N seconds, reconnect
    ) -> None:
        self.url = url
        self.recon = reconnection or ReconnectionConfig()
        self.ws_opts = dict(
            ping_interval=ping_interval,
            ping_timeout=ping_timeout,
            max_size=max_size,
            max_queue=max_queue,
            compression=compression,
        )
        self._state = ConnectionState.DISCONNECTED
        self._ws: Optional[ClientConnection] = None
        self._lock = asyncio.Lock()
        self._stop_event = asyncio.Event()

        # subscriptions
        self._subs: dict[str, ChannelSubscription] = {}
        self._subs_lock = asyncio.Lock()

        # tasks & infra
        self._inbox: asyncio.Queue[bytes | str] = asyncio.Queue(maxsize=inbox_size)
        self._reader_task: Optional[asyncio.Task] = None
        self._worker_tasks: list[asyncio.Task] = []
        self._workers_n = max(1, workers)
        self._hb_task: Optional[asyncio.Task] = None
        self._lag_task: Optional[asyncio.Task] = None
        self._on_reconnected: list[Callable[[], Awaitable[None]]] = []

        # metrics
        self._total_messages = 0
        self._last_rx = monotonic()

        # heartbeat config
        self._app_ping_interval = app_ping_interval
        self._no_rx_reconnect_after = no_rx_reconnect_after

        # message rate tracking
        self._message_tracker = MessageRateTracker(max_window_seconds=900, bucket_size=1.0)

        self.log = logger

    # ---------- public props ----------
    @property
    def state(self) -> ConnectionState:
        return self._state

    @property
    def is_connected(self) -> bool:
        return self._state == ConnectionState.CONNECTED and self._ws is not None

    @property
    def subscriptions(self) -> dict[str, ChannelSubscription]:
        """Read-only access to current subscriptions"""
        return self._subs.copy()

    @property
    def total_messages(self) -> int:
        return self._total_messages

    def get_message_count(self, window_seconds: float) -> int:
        return self._message_tracker.get_count(window_seconds)

    def get_message_rate(self, window_seconds: float) -> float:
        return self._message_tracker.get_rate(window_seconds)

    def on_reconnected(self, cb: Callable[[], Awaitable[None]]) -> None:
        self._on_reconnected.append(cb)

    # ---------- lifecycle ----------
    async def connect(self) -> None:
        async with self._lock:
            if self._state in (ConnectionState.CONNECTED, ConnectionState.CONNECTING):
                return
            self._state = ConnectionState.CONNECTING
            self._stop_event.clear()
            try:
                self._ws = await websockets.connect(self.url, **self.ws_opts)  # type: ignore
                self._state = ConnectionState.CONNECTED
                self._start_tasks()
                self.log.info(f"WS connected: {self.url}")
            except Exception as e:
                self._state = ConnectionState.DISCONNECTED
                self.log.error(f"WS connect failed: {e}")
                raise

    async def disconnect(self) -> None:
        async with self._lock:
            if self._state in (ConnectionState.DISCONNECTED, ConnectionState.CLOSED):
                return
            self._state = ConnectionState.CLOSING
            self._stop_event.set()
            await self._stop_tasks()
            if self._ws:
                with contextlib.suppress(Exception):
                    await self._ws.close()
                self._ws = None
            self._state = ConnectionState.CLOSED
            self.log.info("WS disconnected")

    # ---------- subscriptions ----------
    async def subscribe(self, channel: str, handler: Callable[[dict], Awaitable[None]], **params) -> None:
        if not self.is_connected:
            raise ConnectionError("Not connected")
        async with self._subs_lock:
            self._subs[channel] = ChannelSubscription(channel, handler, params)
        await self._send_subscription_message(channel, params)

    async def unsubscribe(self, channel: str) -> None:
        async with self._subs_lock:
            if channel not in self._subs:
                return
            await self._send_unsubscription_message(channel)
            del self._subs[channel]

    # ---------- messaging ----------
    async def send(self, message: dict) -> None:
        if not self._ws:
            raise ConnectionError("Not connected")
        await self._ws.send(json.dumps(message))

    async def wait_until_closed(self) -> None:
        await self._stop_event.wait()

    # ---------- internals ----------
    def _start_tasks(self) -> None:
        # reader
        self._reader_task = asyncio.create_task(self._reader_loop(), name="ws-reader")
        # workers
        self._worker_tasks = [
            asyncio.create_task(self._worker_loop(), name=f"ws-worker-{i}") for i in range(self._workers_n)
        ]
        # heartbeat
        if self._app_ping_interval:
            self._hb_task = asyncio.create_task(self._heartbeat_loop(), name="ws-heartbeat")
        # loop lag
        self._lag_task = asyncio.create_task(self._loop_lag_monitor(), name="ws-looplag")

    async def _stop_tasks(self) -> None:
        logger.debug("Canceling tasks...")
        victims = [self._reader_task, self._hb_task, self._lag_task, *self._worker_tasks]
        # Filter out None values
        valid_tasks = [t for t in victims if t is not None]

        for t in valid_tasks:
            if not t.done():
                t.cancel()

        logger.debug("Waiting for tasks to complete...")
        if valid_tasks:
            results = await asyncio.gather(*valid_tasks, return_exceptions=True)
            for t, r in zip(valid_tasks, results):
                if isinstance(r, Exception) and not isinstance(r, asyncio.CancelledError):
                    self.log.debug(f"task {t.get_name() if hasattr(t, 'get_name') else t} finished with {r}")

        logger.debug("Tasks canceled and completed")
        self._reader_task = None
        self._hb_task = None
        self._lag_task = None
        self._worker_tasks = []

    async def _reader_loop(self) -> None:
        while not self._stop_event.is_set():
            try:
                if not self._ws:
                    await asyncio.sleep(0.1)
                    continue
                raw = await self._ws.recv()
                self._last_rx = monotonic()
                try:
                    self._inbox.put_nowait(raw)
                except asyncio.QueueFull:
                    # Drop oldest to keep the stream fresh
                    with contextlib.suppress(Exception):
                        _ = self._inbox.get_nowait()
                        self._inbox.task_done()
                    await self._inbox.put(raw)
            except (ConnectionClosedOK, ConnectionClosedError, ConnectionClosed) as e:
                self.log.warning(f"WS closed: code={getattr(e, 'code', None)} reason={getattr(e, 'reason', None)}")
                if not self._stop_event.is_set():
                    asyncio.create_task(self._handle_reconnection())
                break
            except asyncio.CancelledError:
                break
            except Exception as e:
                self.log.error(f"Reader error: {e}")
                if not self._stop_event.is_set():
                    asyncio.create_task(self._handle_reconnection())
                break

    async def _worker_loop(self) -> None:
        while not self._stop_event.is_set():
            try:
                raw = await self._inbox.get()
                self._message_tracker.record_message()

                try:
                    if isinstance(raw, bytes):
                        msg = orjson.loads(raw)
                    else:
                        msg = orjson.loads(raw.encode("utf-8"))
                except orjson.JSONDecodeError as e:
                    self.log.error(f"JSON parse error: {e}")
                    continue

                self._total_messages += 1
                channel = self._extract_channel(msg)
                handled = False
                if channel:
                    sub = self._subs.get(channel)
                    if sub:
                        try:
                            await sub.handler(msg)
                            handled = True
                        except Exception as e:
                            self.log.error(f"Handler error for {channel}: {e}")
                if not handled:
                    await self._handle_unknown_message(msg)
            except Exception as e:
                tb = traceback.format_exc()
                self.log.warning(f"Worker error: {e}\n{tb}")
            finally:
                with contextlib.suppress(Exception):
                    self._inbox.task_done()

    async def _handle_reconnection(self) -> None:
        self.log.debug("Handling reconnection")
        if not self.recon.enabled:
            logger.debug("Reconnection disabled, stopping")
            self._stop_event.set()
            self._state = ConnectionState.DISCONNECTED
            return

        # stop tasks and close old connection
        try:
            logger.debug("Stopping tasks")
            await self._stop_tasks()
        except Exception as e:
            self.log.error(f"Error stopping tasks: {e}")

        if self._ws:
            with contextlib.suppress(Exception):
                await self._ws.close()
            self._ws = None

        self.log.debug("Stopped tasks and closed old connection")

        # backoff
        for attempt in range(1, self.recon.max_retries + 1):
            self._state = ConnectionState.RECONNECTING
            delay = min(self.recon.initial_delay * (self.recon.exponential_base ** (attempt - 1)), self.recon.max_delay)
            delay += random.uniform(-self.recon.jitter, self.recon.jitter)
            delay = max(0.0, delay)
            self.log.info(f"Reconnecting in {delay:.2f}s (attempt {attempt}/{self.recon.max_retries})")
            await asyncio.sleep(delay)
            try:
                self._ws = await websockets.connect(self.url, **self.ws_opts)  # type: ignore
                self._state = ConnectionState.CONNECTED
                self._start_tasks()
                await self._resubscribe_all()
                for cb in self._on_reconnected:
                    with contextlib.suppress(Exception):
                        await cb()
                self.log.info("Reconnected")
                return
            except Exception as e:
                self.log.warning(f"Reconnect attempt {attempt} failed: {e}")
                continue

        self.log.error("Max reconnect attempts reached; giving up.")
        self._stop_event.set()

    async def _resubscribe_all(self) -> None:
        async with self._subs_lock:
            items = list(self._subs.items())
        for channel, sub in items:
            with contextlib.suppress(Exception):
                await self._send_subscription_message(channel, sub.params)
            await asyncio.sleep(0.05)  # gentle throttle

    # ----- hooks for subclasses -----
    async def _send_subscription_message(self, channel: str, params: dict[str, Any]) -> None:
        # Default: {"type":"subscribe","channel":channel,...params}
        await self.send({"type": "subscribe", "channel": channel, **params})

    async def _send_unsubscription_message(self, channel: str) -> None:
        await self.send({"type": "unsubscribe", "channel": channel})

    def _extract_channel(self, message: dict) -> Optional[str]:
        try:
            channel = message.get("channel", None)
            return channel
        except Exception:
            return None

    async def _handle_unknown_message(self, message: dict) -> None:
        # Override for heartbeats/system messages
        pass

    # ----- optional app heartbeat in base -----
    def _app_ping_payload(self) -> Optional[dict]:
        # Subclasses can return e.g. {"type":"ping"}
        return None

    async def _heartbeat_loop(self) -> None:
        interval = float(cast(float, self._app_ping_interval))
        no_rx = float(cast(float, self._no_rx_reconnect_after)) if self._no_rx_reconnect_after else None
        while not self._stop_event.is_set():
            try:
                if self._ws and interval:
                    payload = self._app_ping_payload()
                    if payload:
                        await self.send(payload)
                if no_rx and (monotonic() - self._last_rx) > no_rx:
                    self.log.warning(f"No inbound traffic for {no_rx:.1f}s -> reconnect")
                    asyncio.create_task(self._handle_reconnection())
                    return
            except asyncio.CancelledError:
                return
            except Exception as e:
                self.log.warning(f"heartbeat error: {e}")
            await asyncio.sleep(interval or 1.0)

    async def _loop_lag_monitor(self) -> None:
        loop = asyncio.get_running_loop()
        while not self._stop_event.is_set():
            t = loop.time()
            await asyncio.sleep(1)
            lag_ms = max(0.0, (loop.time() - t - 1.0) * 1000.0)
            if lag_ms > 1000.0:  # 1 second
                self.log.debug(f"Event-loop lag: {lag_ms:.1f} ms")



from typing import Any

from qubx import logger
from qubx.connectors.ccxt.broker import CcxtBroker
from qubx.core.basics import Instrument, OrderSide
from qubx.core.exceptions import BadRequest


class BinanceCcxtBroker(CcxtBroker):
    def __init__(
        self,
        *args,
        enable_price_match: bool = False,
        price_match_ticks: int = 5,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.enable_price_match = enable_price_match
        self.price_match_ticks = price_match_ticks

    def _prepare_order_payload(
        self,
        instrument: Instrument,
        order_side: OrderSide,
        order_type: str,
        amount: float,
        price: float | None = None,
        client_id: str | None = None,
        time_in_force: str = "gtc",
        **options,
    ) -> dict[str, Any]:
        payload = super()._prepare_order_payload(
            instrument, order_side, order_type, amount, price, client_id, time_in_force, **options
        )
        params = payload.get("params", {})
        if "priceMatch" in options:
            params["priceMatch"] = options["priceMatch"]

        quote = self.data_provider.get_quote(instrument)
        if quote is None:
            logger.warning(f"[<y>{instrument.symbol}</y>] :: Quote is not available for order creation.")
            raise BadRequest(f"Quote is not available for price match for {instrument.symbol}")

        if time_in_force == "gtx" and price is not None and self.enable_price_match:
            if (order_side == "BUY" and quote.bid - price < self.price_match_ticks * instrument.tick_size) or (
                order_side == "SELL" and price - quote.ask < self.price_match_ticks * instrument.tick_size
            ):
                params["priceMatch"] = "QUEUE"
                logger.debug(f"[<y>{instrument.symbol}</y>] :: Price match is set to QUEUE. Price will be ignored.")

        if "priceMatch" in params:
            # - if price match is set, we don't need to specify the price
            payload["price"] = None

        payload["params"] = params
        return payload



"""
In-Memory Metric Emitter.

This module provides an implementation of IMetricEmitter that stores metrics in memory
using a pandas DataFrame for easy access and analysis.
"""

from typing import Any, cast

import pandas as pd

from qubx import logger
from qubx.core.basics import Instrument, dt_64
from qubx.core.interfaces import IStrategyContext
from qubx.emitters.base import BaseMetricEmitter
from qubx.utils.ntp import time_now


class InMemoryMetricEmitter(BaseMetricEmitter):
    """
    Emits metrics to an in-memory DataFrame.

    This emitter stores all metrics in a pandas DataFrame for easy access and analysis.
    Useful for testing, debugging, or real-time analysis of strategy metrics.
    """

    def __init__(
        self,
        stats_to_emit: list[str] | None = None,
        stats_interval: str = "1m",
        tags: dict[str, Any] | None = None,
        max_rows: int | None = None,
    ):
        """
        Initialize the In-Memory Metric Emitter.

        Args:
            stats_to_emit: Optional list of specific stats to emit
            stats_interval: Interval for emitting strategy stats (default: "1m")
            tags: Dictionary of default tags/labels to include with all metrics
            max_rows: Maximum number of rows to keep in memory (oldest entries are dropped)
        """
        super().__init__(stats_to_emit, stats_interval, tags)

        # Store emitted rows as a list of dicts for efficiency
        self._rows: list[dict] = []
        self._tag_columns = set()
        self._max_rows = max_rows
        logger.info(f"[InMemoryMetricEmitter] Initialized with max_rows={max_rows}")

    def notify(self, context: IStrategyContext) -> None:
        """
        Notify the metric emitter of a time update.

        This method checks if enough time has passed since the last emission
        and emits metrics if necessary.

        Args:
            context: The strategy context to get statistics from
        """
        # Convert to pandas timestamp for easier time calculations
        current_time = pd.Timestamp(context.time())

        # Floor current time to the stats interval
        floored_current = current_time.floor(self._stats_interval)

        if self._last_emission_time is None:
            self._last_emission_time = floored_current
            return

        floored_last = self._last_emission_time

        # Only emit if we've entered a new interval
        if floored_current > floored_last:
            self.emit_strategy_stats(context)
            self._last_emission_time = floored_current

    def _emit_impl(self, name: str, value: float, tags: dict[str, str], timestamp: dt_64 | None = None) -> None:
        """
        Implementation of emit for in-memory storage.

        Args:
            name: Name of the metric
            value: Value of the metric
            tags: Dictionary of tags/labels for the metric (already merged with default tags)
            timestamp: Optional timestamp for the metric
        """
        try:
            # Use NTP-synchronized time if no timestamp provided
            current_timestamp = timestamp if timestamp is not None else time_now()

            # Convert numpy datetime64 to pandas Timestamp
            if isinstance(current_timestamp, dt_64):
                current_timestamp = pd.Timestamp(current_timestamp)

            # Extract symbol and exchange from tags
            symbol = tags.pop("symbol", None)
            exchange = tags.pop("exchange", None)

            # Create base row data
            row_data = {
                "timestamp": current_timestamp,
                "name": name,
                "value": value,
                "symbol": symbol,
                "exchange": exchange,
            }

            # Add any additional tags as columns
            for tag_key, tag_value in tags.items():
                row_data[tag_key] = tag_value
                self._tag_columns.add(tag_key)

            self._rows.append(row_data)
            # Trim list if max_rows is set
            if self._max_rows is not None and len(self._rows) > self._max_rows:
                self._rows = self._rows[-self._max_rows :]

        except Exception as e:
            logger.error(f"[InMemoryMetricEmitter] Failed to emit metric {name}: {e}")

    def get_dataframe(
        self,
        instrument: Instrument | None = None,
        symbol: str | None = None,
        exchange: str | None = None,
        metric_name: str | None = None,
        start_time: pd.Timestamp | None = None,
        end_time: pd.Timestamp | None = None,
        copy: bool = True,
    ) -> pd.DataFrame:
        """
        Get the metrics DataFrame with optional filtering.
        """
        if not self._rows:
            df = pd.DataFrame(columns=["timestamp", "name", "value", "symbol", "exchange"])
        else:
            df = pd.DataFrame(self._rows.copy())
            # Ensure correct dtypes
            df = df.astype(
                {
                    "timestamp": "datetime64[ns]",
                    "name": "string",
                    "value": "float64",
                    "symbol": "string",
                    "exchange": "string",
                }
            )
        if copy:
            df = df.copy()

        # Filter by instrument (takes precedence over symbol/exchange)
        if instrument is not None:
            df = df[(df["symbol"] == instrument.symbol) & (df["exchange"] == instrument.exchange)]
        else:
            if symbol is not None:
                df = df[df["symbol"] == symbol]
            if exchange is not None:
                df = df[df["exchange"] == exchange]
        if metric_name is not None:
            df = df[df["name"] == metric_name]
        if start_time is not None:
            df = df[df["timestamp"] >= start_time]
        if end_time is not None:
            df = df[df["timestamp"] <= end_time]

        return cast(pd.DataFrame, df)

    def get_latest_metrics(
        self, instrument: Instrument | None = None, symbol: str | None = None, exchange: str | None = None
    ) -> pd.DataFrame:
        """
        Get the latest metrics for each metric name, optionally filtered by instrument.

        Args:
            instrument: Filter by specific instrument
            symbol: Filter by symbol (ignored if instrument is provided)
            exchange: Filter by exchange (ignored if instrument is provided)

        Returns:
            pd.DataFrame: DataFrame with latest metrics
        """
        df = self.get_dataframe(instrument=instrument, symbol=symbol, exchange=exchange, copy=False)

        if df.empty:
            return df

        # Group by metric name and get the latest timestamp for each
        latest_df = df.loc[df.groupby("name")["timestamp"].idxmax()]
        return latest_df.reset_index(drop=True)

    def get_metric_summary(self) -> pd.DataFrame:
        """
        Get a summary of all metrics including count, latest timestamp, and unique instruments.

        Returns:
            pd.DataFrame: Summary statistics for each metric
        """
        if not self._rows:
            return pd.DataFrame(columns=["metric_name", "count", "latest_timestamp", "unique_instruments"])

        summary_data = []

        for metric_name in set(row["name"] for row in self._rows):
            metric_df = [row for row in self._rows if row["name"] == metric_name]

            # Count unique instruments (combinations of symbol and exchange)
            unique_instruments = len(
                set((row["symbol"], row["exchange"]) for row in metric_df if row["symbol"] and row["exchange"])
            )

            summary_data.append(
                {
                    "metric_name": metric_name,
                    "count": len(metric_df),
                    "latest_timestamp": max(row["timestamp"] for row in metric_df),
                    "unique_instruments": unique_instruments,
                }
            )

        return pd.DataFrame(summary_data)

    def clear(self) -> None:
        """Clear all stored metrics."""
        self._rows.clear()
        logger.info("[InMemoryMetricEmitter] Cleared all stored metrics")

    def get_instruments(self) -> list[tuple[str, str]]:
        """
        Get all unique instruments (symbol, exchange pairs) in the stored metrics.

        Returns:
            list[tuple[str, str]]: List of (symbol, exchange) tuples
        """
        return [(row["symbol"], row["exchange"]) for row in self._rows if row["symbol"] and row["exchange"]]

    @property
    def shape(self) -> tuple[int, int]:
        """Get the shape of the stored DataFrame."""
        return self.get_dataframe(copy=False).shape

    @property
    def memory_usage(self) -> pd.Series:
        """Get memory usage of the stored DataFrame."""
        return self.get_dataframe(copy=False).memory_usage(deep=True)



import asyncio
import json
import threading
import urllib.parse
from collections import defaultdict
from typing import Any, Dict, Tuple

import aiohttp
import pandas as pd

from qubx import logger
from qubx.core.basics import (
    CtrlChannel,
    DataType,
    Instrument,
    ITimeProvider,
    dt_64,
)
from qubx.core.interfaces import IDataProvider, IHealthMonitor
from qubx.core.series import Bar, Quote, Trade
from qubx.data.tardis import TARDIS_EXCHANGE_MAPPERS
from qubx.health import DummyHealthMonitor
from qubx.utils.misc import AsyncThreadLoop, synchronized

from .utils import (
    tardis_convert_orderbook,
    tardis_convert_quote,
    tardis_convert_trade,
    tardis_extract_timeframe,
)


class TardisDataProvider(IDataProvider):
    """
    Data provider implementation for Tardis market data service.

    Uses Tardis Machine server for accessing normalized and historical market data
    via WebSocket and HTTP APIs.
    """

    def __init__(
        self,
        host: str,
        port: int,
        exchange: str,
        time_provider: ITimeProvider,
        channel: CtrlChannel,
        health_monitor: IHealthMonitor | None = None,
        asyncio_loop: asyncio.AbstractEventLoop | None = None,
    ):
        """
        Initialize the Tardis data provider.

        Args:
            host: Tardis Machine server host
            port: Tardis Machine server WebSocket port
            exchange: Exchange identifier
            time_provider: Time provider instance
            channel: Control channel for data communication
            health_monitor: Optional health monitoring interface
            asyncio_loop: Optional asyncio event loop. If not provided, a new loop will be created.
        """
        self.time_provider = time_provider
        self.channel = channel
        self._exchange_name = exchange
        self._exchange_id = TARDIS_EXCHANGE_MAPPERS.get(exchange.lower(), exchange.lower())
        self._host = host
        self._port = port
        self._health_monitor = health_monitor or DummyHealthMonitor()

        # Use provided loop or create a new one
        self._own_loop = asyncio_loop is None

        if self._own_loop:
            # Create a new loop and start it in a daemon thread
            self._event_loop = asyncio.new_event_loop()
            self._thread = threading.Thread(
                target=self._run_event_loop, daemon=True, name=f"tardis-{exchange.lower()}-loop"
            )
            self._thread.start()
            self._loop = AsyncThreadLoop(self._event_loop)
        else:
            # Use the provided loop
            # asyncio_loop won't be None here since we check with self._own_loop
            assert asyncio_loop is not None
            self._event_loop = asyncio_loop
            self._thread = None
            self._loop = AsyncThreadLoop(asyncio_loop)

        # Websocket base URLs
        self._ws_url_base = f"ws://{host}:{port}"
        self._http_url_base = f"http://{host}:{port - 1}"

        # Subscriptions tracking
        self._subscriptions: defaultdict[str, set[Instrument]] = defaultdict(set)
        self._last_quotes: defaultdict[Instrument, Quote | None] = defaultdict(lambda: None)

        # Store subscription parameters
        self._subscription_params: Dict[Tuple[str, Instrument], Dict[str, Any]] = {}

        # Symbol to instrument mapping for quick lookups
        self._symbol_to_instrument: dict[str, Instrument] = {}

        # WebSocket connection
        self._ws = None
        self._ws_task = None

        logger.info(f"{self.__prefix()} Initialized Tardis Data Provider")

    def __prefix(self) -> str:
        """Create a colored prefix with the exchange name for logging."""
        return f"<yellow>{self._exchange_name}</yellow>"

    def _run_event_loop(self):
        """Run the event loop in a separate thread."""
        # This is only called when we have a valid event loop
        asyncio.set_event_loop(self._event_loop)
        self._event_loop.run_forever()

    @property
    def is_simulation(self) -> bool:
        """Check if data provider is in simulation mode."""
        return False

    def is_connected(self) -> bool:
        """
        Check if the data provider is currently connected to the exchange.

        Returns:
            bool: True if WebSocket is connected, False otherwise
        """
        return self._ws is not None and not self._ws.closed

    def subscribe(
        self,
        subscription_type: str,
        instruments: set[Instrument],
        reset: bool = False,
    ) -> None:
        """Subscribe to market data for a list of instruments."""
        sub_type, params = DataType.from_str(subscription_type)

        if reset:
            self._subscriptions[sub_type] = set(instruments)
        else:
            self._subscriptions[sub_type].update(instruments)

        # Store subscription parameters for each instrument
        for instrument in instruments:
            self._subscription_params[(sub_type, instrument)] = params

        # Stop previous connection if it's running
        if self._ws_task is not None and not self._ws_task.done():
            try:
                self._loop.submit(self._close_websocket_connection()).result(timeout=5)
            except TimeoutError:
                logger.warning(f"{self.__prefix()} Timeout while closing previous WebSocket connection")
            except Exception as e:
                logger.error(f"{self.__prefix()} Error while closing previous WebSocket connection: {e}")
            self._ws_task = None

        # Start a new WebSocket connection with updated subscriptions
        self._ws_task = self._loop.submit(self._start_websocket_connection())

    def unsubscribe(self, subscription_type: str | None, instruments: set[Instrument]) -> None:
        """Unsubscribe from market data for a list of instruments."""
        if subscription_type is None:
            # Unsubscribe from all subscription types
            for sub_type in list(self._subscriptions.keys()):
                self._subscriptions[sub_type] = self._subscriptions[sub_type].difference(instruments)
                # Remove subscription parameters
                for instrument in instruments:
                    if (sub_type, instrument) in self._subscription_params:
                        del self._subscription_params[(sub_type, instrument)]
        else:
            sub_type, _ = DataType.from_str(subscription_type)
            if sub_type in self._subscriptions:
                self._subscriptions[sub_type] = self._subscriptions[sub_type].difference(instruments)
                # Remove subscription parameters
                for instrument in instruments:
                    if (sub_type, instrument) in self._subscription_params:
                        del self._subscription_params[(sub_type, instrument)]

        # Stop previous connection if it's running
        if self._ws_task is not None and not self._ws_task.done():
            try:
                self._loop.submit(self._close_websocket_connection()).result(timeout=5)
            except TimeoutError:
                logger.warning(f"{self.__prefix()} Timeout while closing WebSocket connection")
            except Exception as e:
                logger.error(f"{self.__prefix()} Error while closing WebSocket connection: {e}")
            self._ws_task = None

        # Start a new WebSocket connection with updated subscriptions
        self._ws_task = self._loop.submit(self._start_websocket_connection())

    def has_subscription(self, instrument: Instrument, subscription_type: str) -> bool:
        """Check if an instrument has a subscription."""
        sub_type, _ = DataType.from_str(subscription_type)
        return sub_type in self._subscriptions and instrument in self._subscriptions[sub_type]

    def get_subscriptions(self, instrument: Instrument | None = None) -> list[str]:
        """Get all subscriptions for an instrument."""
        if instrument is not None:
            return [sub for sub, instrs in self._subscriptions.items() if instrument in instrs]
        return list(self._subscriptions.keys())

    def get_subscribed_instruments(self, subscription_type: str | None = None) -> list[Instrument]:
        """Get a list of instruments that are subscribed to a specific subscription type."""
        if subscription_type is None:
            # Return all subscribed instruments across all subscription types
            result = set()
            for instruments in self._subscriptions.values():
                result.update(instruments)
            return list(result)

        sub_type, _ = DataType.from_str(subscription_type)
        return list(self._subscriptions[sub_type]) if sub_type in self._subscriptions else []

    def warmup(self, configs: dict[tuple[str, Instrument], str]) -> None:
        """Run warmup for subscriptions."""
        warmup_tasks = []

        for (sub_type, instrument), period in configs.items():
            data_type, params = DataType.from_str(sub_type)

            if data_type == DataType.OHLC:
                warmup_tasks.append(self._warmup_ohlc(instrument, period, params.get("timeframe", "1m")))
            elif data_type == DataType.TRADE:
                warmup_tasks.append(self._warmup_trades(instrument, period))
            elif data_type == DataType.ORDERBOOK:
                warmup_tasks.append(
                    self._warmup_orderbook(
                        instrument, period, params.get("depth", 10), params.get("tick_size_pct", 0.01)
                    )
                )

        if warmup_tasks:
            # Run warmup tasks
            async def run_warmup_tasks():
                return await asyncio.gather(*warmup_tasks, return_exceptions=True)

            results = self._loop.submit(run_warmup_tasks()).result(timeout=120)

            # Log any errors
            for i, result in enumerate(results):
                if isinstance(result, Exception):
                    logger.error(f"{self.__prefix()} Warmup task {i} failed: {result}")

    def get_ohlc(self, instrument: Instrument, timeframe: str, nbarsback: int) -> list[Bar]:
        """Get historical OHLC data for an instrument."""
        if nbarsback <= 0:
            return []

        # Calculate start time based on timeframe and nbarsback
        end_time = self.time_provider.time()
        start_time = end_time - pd.Timedelta(timeframe) * nbarsback

        # Format dates for Tardis API
        start_str = pd.Timestamp(start_time).strftime("%Y-%m-%d")
        end_str = pd.Timestamp(end_time).strftime("%Y-%m-%d")

        # Get Tardis-compatible symbol
        symbol = self._get_tardis_symbol(instrument)

        # Build options for HTTP request
        options = {
            "exchange": self._exchange_id,
            "symbols": [symbol],
            "from": start_str,
            "to": end_str,
            "dataTypes": [f"trade_bar_{self._map_timeframe_to_tardis(timeframe)}"],
        }

        encoded_options = urllib.parse.quote_plus(json.dumps(options))
        url = f"{self._http_url_base}/replay-normalized?options={encoded_options}"

        async def fetch_ohlc():
            async with aiohttp.ClientSession() as session:
                async with session.get(url) as response:
                    if response.status != 200:
                        error_text = await response.text()
                        raise ValueError(f"Failed to fetch OHLC data: {error_text}")

                    bars = []
                    async for line in response.content:
                        if not line.strip():
                            continue
                        data = json.loads(line)
                        if data["type"] == "trade_bar":
                            # Convert Tardis trade_bar to Bar object
                            bar_time = pd.Timestamp(data["timestamp"]).value
                            bar = Bar(
                                bar_time,
                                data["open"],
                                data["high"],
                                data["low"],
                                data["close"],
                                volume=data["volume"],
                                bought_volume=data.get("buyVolume", 0),
                            )
                            bars.append(bar)

                    return bars[-nbarsback:] if len(bars) > nbarsback else bars

        try:
            return self._loop.submit(fetch_ohlc()).result(timeout=30)
        except Exception as e:
            logger.error(f"{self.__prefix()} Error fetching OHLC data: {e}")
            return []

    def get_quote(self, instrument: Instrument) -> Quote | None:
        """Get the latest quote for an instrument."""
        return self._last_quotes[instrument]

    def close(self):
        """Close the data provider."""
        # Cancel any ongoing WebSocket tasks
        if self._ws_task is not None and not self._ws_task.done():
            try:
                self._loop.submit(self._close_websocket_connection()).result(timeout=5)
            except TimeoutError:
                logger.warning(f"{self.__prefix()} Timeout while closing WebSocket connection")
            except Exception as e:
                logger.error(f"{self.__prefix()} Error while closing WebSocket connection: {e}")
            finally:
                self._ws_task = None
                self._ws = None

        # If we created our own loop, stop it
        if self._own_loop and self._thread is not None:
            try:
                # Stop the event loop
                # event_loop is guaranteed to be non-None if _own_loop is True
                self._event_loop.stop()

                # Wait for the thread to join
                if self._thread.is_alive():
                    self._thread.join(timeout=5)
            except Exception as e:
                logger.error(f"{self.__prefix()} Error shutting down event loop: {e}")

        logger.info(f"{self.__prefix()} Tardis data provider closed")

    def exchange(self) -> str:
        """Return the name of the exchange this provider reads data."""
        return self._exchange_name

    def _get_tardis_symbol(self, instrument: Instrument) -> str:
        """
        Get the symbol in Tardis-compatible format.

        For bitfinex-derivatives, strip the 't' prefix if it exists.
        Also stores the mapping from symbols to instruments for quick lookups.
        """
        symbol = instrument.exchange_symbol
        if self._exchange_id == "bitfinex-derivatives" and symbol.startswith("t"):
            tardis_symbol = symbol[1:]
        else:
            tardis_symbol = symbol

        # Store the mapping from Tardis symbol to instrument
        self._symbol_to_instrument[tardis_symbol] = instrument
        return tardis_symbol

    @synchronized
    async def _start_websocket_connection(self):
        """Start the WebSocket connection to Tardis Machine."""
        if not self._subscriptions:
            logger.debug(f"{self.__prefix()} No active subscriptions, not starting WebSocket")
            return

        # Build options for the WebSocket connection
        stream_options = []

        # Map all instruments by data type
        all_instruments = set()
        for instruments in self._subscriptions.values():
            all_instruments.update(instruments)

        # Use exchange_symbol instead of symbol, and handle special cases for bitfinex
        symbols = [self._get_tardis_symbol(instrument) for instrument in all_instruments]

        # Map data types to Tardis data types
        data_types = []
        for sub_type in self._subscriptions:
            tardis_type = self._map_data_type_to_tardis(sub_type)
            if tardis_type:
                data_types.append(tardis_type)

        stream_options.append(
            {
                "exchange": self._exchange_id,
                "symbols": symbols,
                "dataTypes": data_types,
            }
        )

        options = urllib.parse.quote_plus(json.dumps(stream_options))
        ws_url = f"{self._ws_url_base}/ws-stream-normalized?options={options}"

        logger.info(
            f"{self.__prefix()} Starting WebSocket connection to Tardis Machine for data types {data_types} symbols {symbols}"
        )

        try:
            async with aiohttp.ClientSession() as session:
                async with session.ws_connect(ws_url) as websocket:
                    self._ws = websocket

                    logger.info(f"{self.__prefix()} WebSocket connected")

                    async for msg in websocket:
                        if msg.type == aiohttp.WSMsgType.TEXT:
                            try:
                                data = json.loads(msg.data)
                                await self._process_tardis_message(data)
                            except json.JSONDecodeError:
                                logger.error(f"{self.__prefix()} Invalid JSON: {msg.data}")
                            except Exception as e:
                                logger.error(f"{self.__prefix()} Error processing message: {e}")
                        elif msg.type == aiohttp.WSMsgType.ERROR:
                            logger.error(f"{self.__prefix()} WebSocket error: {msg.data}")
                            break
                        elif msg.type == aiohttp.WSMsgType.CLOSED:
                            logger.info(f"{self.__prefix()} WebSocket closed")
                            break
        except asyncio.CancelledError:
            logger.info(f"{self.__prefix()} WebSocket connection cancelled")
        except Exception as e:
            logger.error(f"{self.__prefix()} WebSocket error: {e}")
        finally:
            logger.info(f"{self.__prefix()} WebSocket connection closed")
            self._ws = None

    async def _close_websocket_connection(self):
        """Close the WebSocket connection."""
        if self._ws is not None:
            try:
                await self._ws.close()
            finally:
                self._ws = None

    async def _process_tardis_message(self, data: dict[str, Any]):
        """Process a message from Tardis Machine."""
        if not data or "type" not in data or "exchange" not in data:
            return

        # Find matching instrument
        symbol = data.get("symbol")
        if not symbol:
            return

        # Look up the instrument directly from our symbol mapping
        instrument = self._symbol_to_instrument.get(symbol)
        if not instrument:
            # No matching instrument found
            return

        # Get message timestamp
        timestamp = pd.Timestamp(data.get("timestamp", data.get("localTimestamp")))
        msg_time = timestamp.value

        # Record data arrival for health monitoring
        tardis_type = data["type"]
        tardis_name = data["name"] if "name" in data else ""

        if tardis_type == "trade":
            if DataType.TRADE in self._subscriptions and instrument in self._subscriptions[DataType.TRADE]:
                trade = tardis_convert_trade(data, instrument)
                if trade:
                    self.channel.send((instrument, DataType.TRADE, trade, False))

        elif tardis_type == "book_change":
            if DataType.ORDERBOOK in self._subscriptions and instrument in self._subscriptions[DataType.ORDERBOOK]:
                # For book changes, we'd need to maintain a full order book locally
                # This is a simplified implementation
                pass

        elif tardis_type == "quote" or tardis_name == "quote":
            if DataType.QUOTE in self._subscriptions and instrument in self._subscriptions[DataType.QUOTE]:
                quote = tardis_convert_quote(data, instrument)
                if quote:
                    self._last_quotes[instrument] = quote
                    self.channel.send((instrument, DataType.QUOTE, quote, False))

        elif tardis_type == "book_snapshot":
            if DataType.ORDERBOOK in self._subscriptions and instrument in self._subscriptions[DataType.ORDERBOOK]:
                # Get orderbook parameters from subscription
                params = self._get_orderbook_params(instrument)
                levels = params.get("depth", 50)
                tick_size_pct = params.get("tick_size_pct", 0.01)

                orderbook = tardis_convert_orderbook(data, instrument, levels, tick_size_pct)
                if orderbook:
                    # Update last quote from order book
                    self._last_quotes[instrument] = orderbook.to_quote()
                    self.channel.send((instrument, DataType.ORDERBOOK, orderbook, False))

        elif tardis_type == "trade_bar":
            # Map trade_bar to OHLC data type with timeframe
            timeframe = tardis_extract_timeframe(data["name"])
            if timeframe:
                ohlc_type = DataType.OHLC[timeframe]
                sub_type, _ = DataType.from_str(ohlc_type)

                if sub_type in self._subscriptions and instrument in self._subscriptions[sub_type]:
                    bar = Bar(
                        msg_time,
                        data["open"],
                        data["high"],
                        data["low"],
                        data["close"],
                        volume=data["volume"],
                        bought_volume=data.get("buyVolume", 0),
                    )

                    self.channel.send((instrument, ohlc_type, bar, False))

    def _get_orderbook_params(self, instrument: Instrument) -> Dict[str, Any]:
        """
        Get orderbook parameters for a specific instrument.
        Returns default parameters if not found.
        """
        key = (DataType.ORDERBOOK, instrument)
        if key in self._subscription_params:
            return self._subscription_params[key]
        return {"depth": 50, "tick_size_pct": 0.01}  # Default parameters

    async def _warmup_ohlc(self, instrument: Instrument, period: str, timeframe: str):
        """Fetch historical OHLC data for warmup."""
        # Calculate start and end time
        end_time = self.time_provider.time()
        start_time = end_time - pd.Timedelta(period)

        # Format dates for Tardis API
        start_str = pd.Timestamp(start_time).strftime("%Y-%m-%d")
        end_str = pd.Timestamp(end_time).strftime("%Y-%m-%d")

        # Get Tardis-compatible symbol
        symbol = self._get_tardis_symbol(instrument)

        # Build options for HTTP request
        options = {
            "exchange": self._exchange_id,
            "symbols": [symbol],
            "from": start_str,
            "to": end_str,
            "dataTypes": [f"trade_bar_{self._map_timeframe_to_tardis(timeframe)}"],
        }

        encoded_options = urllib.parse.quote_plus(json.dumps(options))
        url = f"{self._http_url_base}/replay-normalized?options={encoded_options}"

        logger.info(f"{self.__prefix()} Warming up OHLC {timeframe} for {symbol} over {period}")

        async with aiohttp.ClientSession() as session:
            async with session.get(url) as response:
                if response.status != 200:
                    error_text = await response.text()
                    raise ValueError(f"Failed to fetch warmup OHLC data: {error_text}")

                bars = []
                async for line in response.content:
                    if not line.strip():
                        continue
                    data = json.loads(line)
                    if data["type"] == "trade_bar" and data["symbol"] == symbol:
                        # Convert Tardis trade_bar to Bar object
                        bar_time = pd.Timestamp(data["timestamp"]).value
                        bar = Bar(
                            bar_time,
                            data["open"],
                            data["high"],
                            data["low"],
                            data["close"],
                            volume=data["volume"],
                            bought_volume=data.get("buyVolume", 0),
                        )
                        bars.append(bar)

                # Send all bars as historical data
                if bars:
                    self.channel.send((instrument, DataType.OHLC[timeframe], bars, True))
                    logger.info(f"{self.__prefix()} Loaded {len(bars)} {timeframe} bars for {symbol}")

    async def _warmup_trades(self, instrument: Instrument, period: str):
        """Fetch historical trade data for warmup."""
        # Calculate start and end time
        end_time = self.time_provider.time()
        start_time = end_time - pd.Timedelta(period)

        # Format dates for Tardis API
        start_str = pd.Timestamp(start_time).strftime("%Y-%m-%d")
        end_str = pd.Timestamp(end_time).strftime("%Y-%m-%d")

        # Get Tardis-compatible symbol
        symbol = self._get_tardis_symbol(instrument)

        # Build options for HTTP request
        options = {
            "exchange": self._exchange_id,
            "symbols": [symbol],
            "from": start_str,
            "to": end_str,
            "dataTypes": ["trade"],
        }

        encoded_options = urllib.parse.quote_plus(json.dumps(options))
        url = f"{self._http_url_base}/replay-normalized?options={encoded_options}"

        logger.info(f"{self.__prefix()} Warming up trades for {symbol} over {period}")

        async with aiohttp.ClientSession() as session:
            async with session.get(url) as response:
                if response.status != 200:
                    error_text = await response.text()
                    raise ValueError(f"Failed to fetch warmup trade data: {error_text}")

                trades = []
                async for line in response.content:
                    if not line.strip():
                        continue
                    data = json.loads(line)
                    if data["type"] == "trade" and data["symbol"] == symbol:
                        # Convert Tardis trade to Trade object
                        trade_time = pd.Timestamp(data["timestamp"]).value
                        trade = Trade(
                            trade_time,
                            data["price"],
                            data["amount"],
                            1 if data.get("side") == "buy" else -1,
                        )
                        trades.append(trade)

                # Send all trades as historical data
                if trades:
                    self.channel.send((instrument, DataType.TRADE, trades, True))
                    logger.info(f"{self.__prefix()} Loaded {len(trades)} trades for {symbol}")

    async def _warmup_orderbook(self, instrument: Instrument, period: str, depth: int, tick_size_pct: float):
        """Fetch historical orderbook data for warmup."""
        # Calculate start and end time
        end_time = self.time_provider.time()
        start_time = end_time - pd.Timedelta(period)

        # Format dates for Tardis API
        start_str = pd.Timestamp(start_time).strftime("%Y-%m-%d")
        end_str = pd.Timestamp(end_time).strftime("%Y-%m-%d")

        # Get Tardis-compatible symbol
        symbol = self._get_tardis_symbol(instrument)

        # Build options for HTTP request
        options = {
            "exchange": self._exchange_id,
            "symbols": [symbol],
            "from": start_str,
            "to": end_str,
            "dataTypes": [f"book_snapshot_{depth}_100ms"],
        }

        encoded_options = urllib.parse.quote_plus(json.dumps(options))
        url = f"{self._http_url_base}/replay-normalized?options={encoded_options}"

        logger.info(f"{self.__prefix()} Warming up orderbook for {symbol} over {period}")

        async with aiohttp.ClientSession() as session:
            async with session.get(url) as response:
                if response.status != 200:
                    error_text = await response.text()
                    raise ValueError(f"Failed to fetch warmup orderbook data: {error_text}")

                orderbooks = []
                async for line in response.content:
                    if not line.strip():
                        continue
                    data = json.loads(line)
                    if data["type"] == "book_snapshot" and data["symbol"] == symbol:
                        # Use the utility function to convert Tardis book_snapshot to OrderBook object
                        orderbook = tardis_convert_orderbook(data, instrument, depth, tick_size_pct)
                        if orderbook:
                            orderbooks.append(orderbook)

                # Send orderbooks as historical data
                if orderbooks:
                    params = {"tick_size_pct": tick_size_pct, "depth": depth}
                    sub_type = f"{DataType.ORDERBOOK}({tick_size_pct}, {depth})"
                    self.channel.send((instrument, sub_type, orderbooks, True))
                    logger.info(f"{self.__prefix()} Loaded {len(orderbooks)} orderbooks for {symbol}")

    def _map_data_type_to_tardis(self, data_type: str) -> str:
        """Map QubX data type to Tardis data type."""
        if data_type == DataType.TRADE:
            return "trade"
        elif data_type == DataType.QUOTE:
            return "quote"
        elif data_type == DataType.ORDERBOOK:
            return "book_snapshot_2000_100ms"  # Default depth and interval
        elif data_type == DataType.OHLC:
            return "trade_bar_1m"  # Default timeframe
        else:
            return ""

    def _map_tardis_type_to_data_type(self, tardis_type: str) -> str:
        """Map Tardis data type to QubX data type."""
        if tardis_type == "trade":
            return DataType.TRADE
        elif tardis_type == "quote":
            return DataType.QUOTE
        elif tardis_type.startswith("book_snapshot"):
            return DataType.ORDERBOOK
        elif tardis_type.startswith("trade_bar"):
            return DataType.OHLC
        else:
            return ""

    def _map_timeframe_to_tardis(self, timeframe: str) -> str:
        """Map QubX timeframe to Tardis timeframe format."""
        # Map common timeframes (1m, 5m, 1h, etc.) to milliseconds format
        tf_map = {
            "1m": "60000ms",
            "5m": "300000ms",
            "15m": "900000ms",
            "30m": "1800000ms",
            "1h": "3600000ms",
            "4h": "14400000ms",
            "1d": "86400000ms",
        }

        # Try direct mapping first
        if timeframe.lower() in tf_map:
            return tf_map[timeframe.lower()]

        # Otherwise, try to parse the timeframe
        try:
            pd_timedelta = pd.Timedelta(timeframe)
            return f"{int(pd_timedelta.total_seconds() * 1000)}ms"
        except:
            # Fallback to default 1m if parsing fails
            logger.warning(f"{self.__prefix()} Could not map timeframe {timeframe}, using 1m")
            return "60000ms"



"""
ExchangeManager: Transparent wrapper for CCXT exchanges with automatic recreation.

Provides seamless exchange recreation without affecting consuming components.
"""

import asyncio
import threading
import time
from typing import Any, Callable, Optional

import pandas as pd

import ccxt.pro as cxp
from qubx import logger
from qubx.core.interfaces import IHealthMonitor, ITimeProvider

# Constants for better maintainability
DEFAULT_CHECK_INTERVAL_SECONDS = 60.0
SECONDS_PER_HOUR = 3600

# Custom stale detection thresholds (in seconds)
STALE_THRESHOLDS = {
    "funding_payment": 12 * SECONDS_PER_HOUR,  # 12 hours = 43,200s
    "open_interest": 30 * 60,  # 30 minutes = 1,800s
    "orderbook": 5 * 60,  # 5 minutes = 300s
    "trade": 60 * 60,  # 60 minutes = 3,600s
    "liquidation": 7 * 24 * SECONDS_PER_HOUR,  # 7 days = 604,800s
    "ohlc": 5 * 60,  # 5 minutes = 300s
    "quote": 2 * 60,  # 2 minutes = 120s
}
DEFAULT_STALE_THRESHOLD_SECONDS = 2 * SECONDS_PER_HOUR  # 2 hours = 7,200s


class ExchangeManager:
    """
    Wrapper for CCXT Exchange that handles recreation internally with self-monitoring.

    Exposes the underlying exchange via .exchange property for explicit access.
    Self-monitors for data stale and triggers recreation automatically.

    Key Features:
    - Explicit .exchange property for CCXT access
    - Self-contained stall detection and recreation triggering
    - Automatic recreation without limits when data stalls
    - Atomic exchange transitions during recreation
    - Background monitoring thread for stall detection
    """

    _exchange: cxp.Exchange  # Type hint that this is always a valid exchange

    def __init__(
        self,
        exchange_name: str,
        factory_params: dict[str, Any],
        health_monitor: IHealthMonitor,
        time_provider: ITimeProvider,
        initial_exchange: Optional[cxp.Exchange] = None,
        check_interval_seconds: float = DEFAULT_CHECK_INTERVAL_SECONDS,
    ):
        """Initialize ExchangeManager with underlying CCXT exchange.

        Args:
            exchange_name: Exchange name for factory (e.g., "binance.um")
            factory_params: Parameters for get_ccxt_exchange()
            initial_exchange: Pre-created exchange instance (from factory)
            check_interval_seconds: How often to check for stale data (default: 60.0)
        """
        self._exchange_name = exchange_name
        self._factory_params = factory_params.copy()
        self._health_monitor = health_monitor
        self._time_provider = time_provider

        # Recreation state
        self._recreation_lock = threading.RLock()
        self._recreation_count = 0  # Track for logging purposes only

        # Stale data detection state
        self._check_interval = check_interval_seconds
        self._last_data_times: dict[str, float] = {}
        self._data_lock = threading.RLock()

        # Monitoring control
        self._monitoring_enabled = False
        self._monitor_thread = None

        # Recreation callback management
        self._recreation_callbacks: list[Callable[[], None]] = []

        # Use provided exchange or create new one
        if initial_exchange:
            self._exchange = initial_exchange
            # Setup exception handler on provided exchange
            self._setup_ccxt_exception_handler(self._exchange)
        else:
            self._exchange = self._create_exchange()

    def _create_exchange(self) -> cxp.Exchange:
        """Create new raw CCXT exchange instance using factory method."""
        try:
            # Import here to avoid circular import (factory → broker → exchange_manager)
            from .factory import get_ccxt_exchange

            # Create raw exchange using factory logic
            ccxt_exchange = get_ccxt_exchange(**self._factory_params)

            # Setup exception handler for the new exchange
            self._setup_ccxt_exception_handler(ccxt_exchange)

            logger.debug(f"Created new {self._exchange_name} exchange instance")
            return ccxt_exchange

        except Exception as e:
            logger.error(f"Failed to create {self._exchange_name} exchange: {e}")
            raise RuntimeError(f"Failed to create {self._exchange_name} exchange: {e}") from e

    def register_recreation_callback(self, callback: Callable[[], None]) -> None:
        """Register callback to be called after successful exchange recreation.

        Args:
            callback: Function to call after successful recreation (no parameters)
        """
        self._recreation_callbacks.append(callback)
        logger.debug(f"Registered recreation callback for {self._exchange_name}")

    def _call_recreation_callbacks(self) -> None:
        """Call all registered recreation callbacks after successful exchange recreation."""
        logger.debug(f"Calling {len(self._recreation_callbacks)} recreation callbacks for {self._exchange_name}")

        for callback in self._recreation_callbacks:
            try:
                callback()
            except Exception as e:
                logger.error(f"Error in recreation callback for {self._exchange_name}: {e}")
                # Continue with other callbacks even if one fails

    def force_recreation(self) -> bool:
        """
        Force recreation due to stale data.

        Returns:
            True if recreation successful, False if failed
        """
        with self._recreation_lock:
            logger.info(f"Stale-triggered recreation for {self._exchange_name}")
            return self._recreate_exchange()

    def _recreate_exchange(self) -> bool:
        """Recreate the underlying exchange (must be called with _recreation_lock held)."""
        self._recreation_count += 1
        logger.warning(f"Recreating {self._exchange_name} exchange (attempt {self._recreation_count})")

        # Create new exchange
        try:
            new_exchange = self._create_exchange()
        except Exception as e:
            logger.error(f"Failed to recreate {self._exchange_name} exchange: {e}")
            return False

        # Atomically replace the exchange
        old_exchange = self._exchange
        self._exchange = new_exchange

        # Clean up old exchange
        try:
            if hasattr(old_exchange, "close") and hasattr(old_exchange, "asyncio_loop"):
                old_exchange.asyncio_loop.call_soon_threadsafe(lambda: asyncio.create_task(old_exchange.close()))
        except Exception as e:
            logger.warning(f"Error closing old {self._exchange_name} exchange: {e}")

        logger.info(f"Successfully recreated {self._exchange_name} exchange")

        # Call recreation callbacks after successful recreation
        self._call_recreation_callbacks()

        return True

    def _extract_ohlc_timeframe(self, event_type: str) -> Optional[str]:
        """Extract timeframe from OHLC event type like 'ohlc(1m)' -> '1m'."""
        if event_type.startswith("ohlc(") and event_type.endswith(")"):
            return event_type[5:-1]  # Simple slice: ohlc(1m) -> 1m
        return None

    def _timeframe_to_seconds(self, timeframe: str) -> int:
        """Convert timeframe string to seconds using pandas.Timedelta."""
        return int(pd.Timedelta(timeframe).total_seconds())

    def _get_stale_threshold(self, event_type: str) -> float:
        """Get stale threshold for specific event type.

        Extracts base data type from parameterized types like 'ohlc(1m)' -> 'ohlc'.
        """
        # Extract base data type (everything before first '(' if present)
        base_event_type = event_type.split("(")[0]
        return float(STALE_THRESHOLDS.get(base_event_type, DEFAULT_STALE_THRESHOLD_SECONDS))

    def start_monitoring(self) -> None:
        """Start background stale data detection monitoring."""
        if self._monitoring_enabled:
            return

        self._monitoring_enabled = True
        self._monitor_thread = threading.Thread(target=self._stale_monitor_loop, daemon=True)
        self._monitor_thread.start()
        logger.debug(f"ExchangeManager: Started stale data detection monitoring for {self._exchange_name}")

    def stop_monitoring(self) -> None:
        """Stop background stall detection monitoring."""
        self._monitoring_enabled = False
        if self._monitor_thread:
            self._monitor_thread = None
        logger.debug(f"ExchangeManager: Stopped stall monitoring for {self._exchange_name}")

    def _stale_monitor_loop(self) -> None:
        """Background thread that checks for data stalls and triggers self-recreation."""
        while self._monitoring_enabled:
            try:
                self._check_and_handle_stales()
                time.sleep(self._check_interval)
            except Exception as e:
                logger.error(f"Error in ExchangeManager stall detection: {e}")
                time.sleep(self._check_interval)

    def _check_and_handle_stales(self) -> None:
        """Check for stalls using custom thresholds per data type."""
        current_time = self._time_provider.time()
        stale_types = []

        with self._data_lock:
            last_event_times = self._health_monitor.get_last_event_times_by_exchange(self._exchange_name)
            for event_type, last_data_time in last_event_times.items():
                time_since_data = current_time - last_data_time
                # Convert timedelta64 to seconds for comparison
                time_since_seconds = float(time_since_data.astype("timedelta64[ns]").astype(int) / 1e9)
                threshold = self._get_stale_threshold(event_type)

                if time_since_seconds > threshold:
                    stale_types.append((event_type, time_since_seconds))

        if not stale_types:
            return  # No stale data detected

        stale_info = ", ".join([f"{event_type}({int(time_since)}s)" for event_type, time_since in stale_types])
        logger.error(f"Data staleness detected in {self._exchange_name}: {stale_info}")

        try:
            logger.info(f"Self-triggering recreation for {self._exchange_name} due to stale data...")
            if self.force_recreation():
                logger.info(f"Stall-triggered recreation successful for {self._exchange_name}")
            else:
                logger.error(f"Stall-triggered recreation failed for {self._exchange_name}")
        except Exception as e:
            logger.error(f"Error during stall-triggered recreation: {e}")

    def _setup_ccxt_exception_handler(self, exchange: cxp.Exchange) -> None:
        """
        Set up global exception handler for the CCXT async loop to handle unretrieved futures.

        This prevents 'Future exception was never retrieved' warnings from CCXT's internal
        per-symbol futures that complete with UnsubscribeError during resubscription.

        Applied to every newly created exchange (initial and recreated).
        """
        asyncio_loop = exchange.asyncio_loop

        def handle_ccxt_exception(loop, context):
            """Handle unretrieved exceptions from CCXT futures."""
            exception = context.get("exception")

            # Handle expected CCXT UnsubscribeError during resubscription
            if exception and "UnsubscribeError" in str(type(exception)):
                return

            # Handle other CCXT-related exceptions quietly if they're in our exchange context
            if exception and any(
                keyword in str(exception) for keyword in [exchange.id, "ohlcv", "orderbook", "ticker"]
            ):
                return

            # For all other exceptions, use the default handler
            if hasattr(loop, "default_exception_handler"):
                loop.default_exception_handler(context)
            else:
                # Fallback logging if no default handler
                logger.warning(f"Unhandled asyncio exception: {context}")

        # Set the custom exception handler on the CCXT loop
        asyncio_loop.set_exception_handler(handle_ccxt_exception)

    # === Exchange Property Access ===
    # Explicit property to access underlying CCXT exchange

    @property
    def exchange(self) -> cxp.Exchange:
        """Access to the underlying CCXT exchange instance.

        Use this property to call CCXT methods: exchange_manager.exchange.fetch_ticker(symbol)

        Returns:
            The current CCXT exchange instance (may change after recreation)
        """
        return self._exchange



"""
Kernel service manager for persistent strategy kernels.

Manages the lifecycle of Jupyter kernels that run independently of the UI,
allowing multiple UI instances to connect to the same running strategy.
"""

import asyncio
import time
from pathlib import Path

from jupyter_client.manager import AsyncKernelManager

from qubx import logger


class KernelService:
    """Manages persistent kernel subprocesses for strategy execution."""

    _active_kernels: dict[str, tuple[AsyncKernelManager, Path]] = {}

    @classmethod
    async def start(
        cls,
        config_file: Path,
        account_file: Path | None,
        paper: bool,
        restore: bool,
    ) -> str:
        """
        Start a persistent kernel with strategy initialization.

        Args:
            config_file: Path to strategy configuration
            account_file: Path to account configuration (optional)
            paper: Whether to run in paper trading mode
            restore: Whether to restore strategy state

        Returns:
            Path to the kernel connection file
        """
        # Create ~/.qubx/kernels directory for connection files
        kernels_dir = Path.home() / ".qubx" / "kernels"
        kernels_dir.mkdir(parents=True, exist_ok=True)

        # Generate unique connection file name
        timestamp = time.strftime("%Y%m%d_%H%M%S")
        strategy_name = config_file.stem
        conn_file = kernels_dir / f"{strategy_name}_{timestamp}.json"

        # Create kernel manager with custom connection file
        km = AsyncKernelManager(kernel_name="python3", connection_file=str(conn_file))
        await km.start_kernel()

        # Write the connection file (now it will use our custom path)
        km.write_connection_file()

        # Wait for kernel to be ready
        kc = km.client()
        kc.start_channels()
        await kc.wait_for_ready()

        logger.info(f"Kernel started with connection file: {conn_file}")

        # Initialize the strategy in the kernel
        from .textual.init_code import generate_init_code

        init_code = generate_init_code(config_file, account_file, paper, restore)
        msg_id = kc.execute(init_code, silent=False)

        # Wait for initialization to complete before allowing UI connections
        logger.info("Waiting for strategy initialization to complete...")
        timeout = 30.0  # Give it 30 seconds to initialize
        start_time = asyncio.get_event_loop().time()

        init_complete = False
        while asyncio.get_event_loop().time() - start_time < timeout:
            try:
                msg = await asyncio.wait_for(kc.get_iopub_msg(timeout=1.0), timeout=1.0)

                if msg["parent_header"].get("msg_id") == msg_id:
                    msg_type = msg["header"]["msg_type"]

                    if msg_type == "status":
                        execution_state = msg["content"].get("execution_state")
                        if execution_state == "idle":
                            # Init code finished executing
                            init_complete = True
                            break
                    elif msg_type == "error":
                        # Init code had an error
                        error_content = msg["content"]
                        logger.error(f"Strategy initialization failed: {error_content.get('ename')}: {error_content.get('evalue')}")
                        raise RuntimeError(f"Strategy initialization failed: {error_content.get('ename')}")
            except asyncio.TimeoutError:
                continue

        if not init_complete:
            logger.warning("Strategy initialization did not complete within timeout")

        # Stop the client channels (we don't need them, clients will reconnect)
        kc.stop_channels()

        # Store kernel manager reference
        cls._active_kernels[str(conn_file)] = (km, config_file)

        return str(conn_file)

    @classmethod
    async def stop(cls, connection_file: str) -> None:
        """
        Stop a kernel identified by its connection file.

        Args:
            connection_file: Path to the kernel connection file
        """
        if connection_file not in cls._active_kernels:
            logger.warning(f"Kernel not found for connection file: {connection_file}")
            return

        km, _ = cls._active_kernels[connection_file]

        try:
            await km.shutdown_kernel(now=False)
            logger.info(f"Kernel stopped: {connection_file}")
        except Exception as e:
            logger.error(f"Error stopping kernel: {e}")
            # Force kill if graceful shutdown fails
            try:
                await km.shutdown_kernel(now=True)
            except Exception:
                pass

        # Remove from active kernels
        del cls._active_kernels[connection_file]

        # Clean up connection file
        try:
            Path(connection_file).unlink(missing_ok=True)
        except Exception as e:
            logger.warning(f"Could not delete connection file: {e}")

    @classmethod
    async def is_alive(cls, connection_file: str) -> bool:
        """
        Check if a kernel is still alive.

        Args:
            connection_file: Path to the kernel connection file

        Returns:
            True if kernel is alive, False otherwise
        """
        if connection_file not in cls._active_kernels:
            return False

        km, _ = cls._active_kernels[connection_file]

        try:
            return await km.is_alive()
        except Exception:
            return False

    @classmethod
    def list_active(cls) -> list[dict[str, str]]:
        """
        List all active kernel sessions by scanning the kernels directory.

        Returns:
            List of dicts with connection_file and config_file info
        """
        kernels_dir = Path.home() / ".qubx" / "kernels"
        if not kernels_dir.exists():
            return []

        result = []
        for conn_file in kernels_dir.glob("*.json"):
            # Parse the filename to extract config info
            # Format: {strategy_name}_{timestamp}.json
            strategy_name = conn_file.stem.rsplit("_", 2)[0] if "_" in conn_file.stem else conn_file.stem
            result.append({
                "connection_file": str(conn_file),
                "strategy_name": strategy_name,
                "timestamp": conn_file.stat().st_mtime,
            })

        # Sort by timestamp (newest first)
        result.sort(key=lambda x: x["timestamp"], reverse=True)
        return result

    @classmethod
    async def cleanup_all(cls) -> None:
        """Stop all active kernels (cleanup on exit)."""
        for conn_file in list(cls._active_kernels.keys()):
            await cls.stop(conn_file)


from typing import Any, Iterator, TypeVar, overload

import numpy as np
from numpy.typing import ArrayLike, DTypeLike, NDArray

_T = TypeVar("_T", bound=np.generic)

class RingBuffer:
    """
    A circular buffer implementation using numpy arrays.

    Provides efficient O(1) append/pop operations at both ends of the buffer.
    When the buffer is full, new elements overwrite the oldest elements.
    """

    def __init__(self, capacity: int, dtype: DTypeLike = ...) -> None:
        """
        Initializes the RingBuffer with a specified capacity and data type.

        Args:
            capacity: The maximum number of elements the buffer can hold.
            dtype: Desired data-type for the array. Default is np.float64.

        Raises:
            ValueError: If the capacity is not a positive integer.
        """
        ...

    @property
    def is_full(self) -> bool:
        """
        Checks if the buffer is full.

        Returns:
            True if the buffer is full, False otherwise.
        """
        ...

    @property
    def dtype(self) -> np.dtype[Any]:
        """
        Returns the data type of the buffer elements.

        Returns:
            The data type of the buffer elements.
        """
        ...

    @property
    def shape(self) -> tuple[int]:
        """
        Returns the shape of the buffer.

        Returns:
            The shape of the buffer as a single-element tuple.
        """
        ...

    @property
    def maxlen(self) -> int:
        """
        Returns the maximum length of the buffer.

        Returns:
            The maximum number of elements the buffer can hold.
        """
        ...

    def append(self, value: Any) -> None:
        """
        Appends a value to the right end of the buffer.

        If the buffer is full, the leftmost element is removed.

        Args:
            value: The value to append.
        """
        ...

    def appendleft(self, value: Any) -> None:
        """
        Appends a value to the left end of the buffer.

        If the buffer is full, the rightmost element is removed.

        Args:
            value: The value to append.
        """
        ...

    def clear(self) -> None:
        """
        Clears the buffer, removing all elements.
        """
        ...

    def peek(self) -> Any:
        """
        Returns the rightmost element without removing it.

        Returns:
            The rightmost element.

        Raises:
            IndexError: If the buffer is empty.
        """
        ...

    def peekleft(self) -> Any:
        """
        Returns the leftmost element without removing it.

        Returns:
            The leftmost element.

        Raises:
            IndexError: If the buffer is empty.
        """
        ...

    def pop(self) -> Any:
        """
        Removes and returns the rightmost element.

        Returns:
            The rightmost element.

        Raises:
            IndexError: If the buffer is empty.
        """
        ...

    def popleft(self) -> Any:
        """
        Removes and returns the leftmost element.

        Returns:
            The leftmost element.

        Raises:
            IndexError: If the buffer is empty.
        """
        ...

    def __array__(self) -> NDArray[Any]:
        """
        Returns the buffer as a numpy array.

        Returns:
            The buffer as a contiguous numpy array.
        """
        ...

    def __len__(self) -> int:
        """
        Returns the number of elements in the buffer.

        Returns:
            The number of elements currently in the buffer.
        """
        ...

    @overload
    def __getitem__(self, item: int) -> Any: ...
    @overload
    def __getitem__(self, item: slice) -> NDArray[Any]: ...
    @overload
    def __getitem__(self, item: ArrayLike) -> NDArray[Any]: ...
    def __getitem__(self, item: int | slice | ArrayLike) -> Any | NDArray[Any]:
        """
        Returns the element(s) at the specified index or slice.

        Args:
            item: The index, slice, or array of indices to retrieve.

        Returns:
            The element(s) at the specified index or slice.
        """
        ...

    def __iter__(self) -> Iterator[Any]:
        """
        Returns an iterator over the buffer.

        Returns:
            An iterator over the buffer elements.
        """
        ...

    def __repr__(self) -> str:
        """
        Returns a string representation of the buffer.

        Returns:
            A string representation of the buffer.
        """
        ...



from typing import cast

import pandas as pd

from qubx import logger
from qubx.core.basics import Instrument
from qubx.core.interfaces import ITimeProvider


class DelistingDetector:
    def __init__(self, time_provider: ITimeProvider, delisting_check_days: int = 1, silent: bool = False):
        self._time_provider = time_provider
        self._delisting_check_days = delisting_check_days
        self._silent = silent

    @property
    def delisting_check_days(self) -> int:
        return self._delisting_check_days

    @delisting_check_days.setter
    def delisting_check_days(self, days: int) -> None:
        self._delisting_check_days = days

    def detect_delistings(self, instruments: list[Instrument]) -> list[Instrument]:
        """
        Detect instruments that are delisting within delisting_check_days.

        Args:
            instruments: List of instruments to check

        Returns:
            List of instruments that will be delisted within the configured time period
        """
        if self._delisting_check_days <= 0:
            return []

        current_time = pd.Timestamp(self._time_provider.time())
        check_ahead = cast(pd.Timestamp, current_time + pd.Timedelta(days=self._delisting_check_days))

        delisting = []
        for instrument in instruments:
            if instrument.delist_date is None:
                continue

            try:
                delist_timestamp = pd.Timestamp(instrument.delist_date).replace(tzinfo=None)
                if bool(pd.isna(delist_timestamp)):
                    continue

                if delist_timestamp <= check_ahead:
                    delisting.append(instrument)
            except Exception as e:
                self._warning(f"Error checking delist date for {instrument}: {e}")
                # On error, exclude from delisting list to be safe
                continue

        return delisting

    def filter_delistings(self, instruments: list[Instrument]) -> list[Instrument]:
        """
        Filter out instruments that are delisting within delisting_check_days.

        This returns instruments that are safe to trade (not delisting soon).

        Args:
            instruments: List of instruments to filter

        Returns:
            Filtered list of instruments without soon-to-delist instruments
        """
        # Reuse detect_delistings to find instruments that should be filtered out
        delisting_instruments = self.detect_delistings(instruments)
        delisting_set = set(delisting_instruments)

        # Log instruments being filtered out
        for instrument in delisting_instruments:
            self._info(f"Filtering out {instrument.symbol} - delisting on {instrument.delist_date}")

        # Return instruments that are NOT in the delisting set
        return [instrument for instrument in instruments if instrument not in delisting_set]

    def _info(self, msg: str) -> None:
        if not self._silent:
            logger.info(f"[DelistingDetector] {msg}")

    def _warning(self, msg: str) -> None:
        if not self._silent:
            logger.warning(f"[DelistingDetector] {msg}")

    def _error(self, msg: str) -> None:
        if not self._silent:
            logger.error(f"[DelistingDetector] {msg}")



import re
from collections import defaultdict
from collections.abc import Callable, Iterator
from dataclasses import dataclass
from functools import wraps

import numpy as np
import pandas as pd
import psycopg as pg

from qubx import logger
from qubx.core.basics import DataType
from qubx.data.containers import RawData, RawMultiData
from qubx.data.registry import storage
from qubx.data.storage import IReader, IStorage, Transformable
from qubx.data.storages.utils import calculate_time_windows_for_chunking, find_column_index_in_list
from qubx.utils.time import handle_start_stop, timedelta_to_str


def _retry(fn):
    @wraps(fn)
    def wrapper(*args, **kw):
        cls = args[0]
        for x in range(cls._reconnect_tries):
            try:
                return fn(*args, **kw)
            except (pg.InterfaceError, pg.OperationalError, AttributeError):
                logger.debug("Database Connection [InterfaceError or OperationalError]")
                cls._connect()

    return wrapper


class PGConnectionHelper:
    _connection: pg.connection.Connection | None
    _reconnect_tries = 5
    _reconnect_idle = 0.1  # wait seconds before retying
    _host: str
    _user: str
    _password: str
    _port: int

    def __init__(self, host="localhost", user="admin", password="quest", port=8812) -> None:
        self._connection = None
        self._user = user
        self._password = password
        self._host = host
        self._port = port
        self._connect()

    def _connect(self):
        self._connection = pg.connect(self.connection_url, autocommit=True)
        logger.debug(f"Connected to QuestDB at {self._host}:{self._port}")

    @property
    def connection_url(self):
        return " ".join(
            [f"user={self._user}", f"password={self._password}", f"host={self._host}", f"port={self._port}"]
        )

    def __getstate__(self):
        if self._connection:
            self._connection.close()
            self._connection = None
        state = self.__dict__.copy()
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        # - don't connect immediately - let _retry decorator handle lazy connection on first use

    def close(self):
        if self._connection:
            self._connection.close()
            self._connection = None
            logger.debug(f"Disconnected from QuestDB at {self._host}:{self._port}")

    @_retry
    def execute(self, query: str) -> tuple[list, list]:
        with self._connection.cursor() as _cursor:  # type: ignore
            _cursor.execute(query)  # type: ignore
            names = [d.name for d in _cursor.description]  # type: ignore
            records = _cursor.fetchall()
            return names, records

    def __del__(self):
        try:
            if self._connection is not None:
                logger.debug("Closing connection")
                self._connection.close()
        except:  # noqa: E722
            pass


@dataclass
class xLTableMetaInfo:
    """
    Table meta info container and decoder
        TODO: we need to fix tables naming in DB to drop any special processing code
    """

    exchange: str
    market_type: str
    dtype: DataType
    data_timeframe: str | None
    table_name: str
    alias_for_record_type: str | None = None

    _TABLES_FIX = {
        ("binance", "umswap"): ("binance.um", "swap"),
        ("binance", "umfutures"): ("binance.um", "future"),
        ("binance", "cmswap"): ("binance.cm", "swap"),
        ("binance", "cmfutures"): ("binance.cm", "future"),
        ("binance", "margin"): ("binance.pm", "swap"),
    }

    _DTYPE_FIX = {
        "candles": "ohlc",
        "orderbooks": "orderbook",
        "liquidations": "aggregated_liquidations",
        # "interest_rates": "record",
    }

    _d_pattern = r"^(.+?)(?:_(\d+(?:min|[mhdw])))?$"

    @staticmethod
    def decode_table_metadata(table_name: str) -> "xLTableMetaInfo | None":
        """
        Decode table name and try to extract metadata:
            binance.umswap.candles_1m -> BINANCE.UM, SWAP, OHLC[1min]
        """
        ss = table_name.split(".")
        if len(ss) > 1:
            exch, mkt, data_type = ss[0], ss[1], (ss[2] if len(ss) > 2 else ss[1])
            exch, mkt = xLTableMetaInfo._TABLES_FIX.get((exch.lower(), mkt.lower()), (exch.lower(), mkt.lower()))

            # - consider it as valid data if we can recognize structure
            if sg := re.match(xLTableMetaInfo._d_pattern, data_type):
                data_type, tframe = sg.groups()
                f_data_type = xLTableMetaInfo._DTYPE_FIX.get(data_type, data_type)
                r_dt, _ = DataType.from_str(f_data_type)
                r_dt = DataType.RECORD if r_dt == DataType.NONE else r_dt
                _alias = data_type if r_dt == DataType.RECORD else None
                return xLTableMetaInfo(exch.upper(), mkt.upper(), r_dt, tframe, table_name, _alias)

        return None


# fmt: off
_ext_frames = pd.to_timedelta(
    [
        "1s", "2s", "3s", "5s", "10s", "15s", "30s",
        "1Min", "2Min", "3Min", "5Min", "10Min", "15Min", "30Min",
        "1h", "2h", "3h", "4h", "6h", "8h", "12h",
        "1d", "1w",
    ]
)
# fmt: on


class QuestDBReader(IReader):
    """
    TODO: create docstring here
    """

    # Info about datatypes and symbols for fast access
    #   | Symbols lookup: {symbol   -> { dtype -> xLTableMetaInfo}
    #   | Dtypes lookup:  {DataType -> ([symbols], xLTableMetaInfo)
    _symbols_lookup: dict[str, dict[DataType, xLTableMetaInfo]]
    _dtype_lookup: dict[str, tuple[set[str], xLTableMetaInfo]]
    _external_sql_builders: dict[str, Callable[[set[str], list[str], str], str]]

    def __init__(
        self,
        exchange: str,
        market: str,
        available_data: list[xLTableMetaInfo],
        pgc: PGConnectionHelper,
        synthetic_ohlc_timeframes_types: bool,
        min_symbols_for_all_data_request: int,
        symbol_column_name="symbol",
    ) -> None:
        self.exchange = exchange
        self.market = market
        self.synthetic_ohlc_timeframes_types = synthetic_ohlc_timeframes_types
        self.min_symbols_for_all_data_request = min_symbols_for_all_data_request
        self.symbol_column_name = symbol_column_name
        self.pgc = pgc
        self._external_sql_builders = {}

        # - build lookup
        self._build_lookups(available_data)

    @staticmethod
    def _convert_time_delta_to_qdb_resample_format(c_tf: str) -> tuple[str, str]:
        """
        Convert standard timedelta format into quest db: "15Min" -> (15, "m) etc
        Returns tuple (number, units)
        """
        if c_tf:
            _t = re.match(r"(\d+)(\w+)", c_tf)
            if _t and len(_t.groups()) > 1:
                number = _t[1]
                units = _t[2][0].lower()
                return number, units
        return c_tf, ""

    def _build_lookups(self, available_data: list[xLTableMetaInfo]):
        """
        TODO: create docstring here
        """
        _symbs_lookup = defaultdict(dict)
        _dtype_lookup = dict()

        for x in available_data:
            # - collect symbols for every table in this reader
            symbols_info = self.pgc.execute(f"select distinct({self.symbol_column_name}) from {x.table_name}")[1]
            symbols, dtypes = [], []
            for s in symbols_info:
                symbols.append(symb := s[0])

                if x.dtype == DataType.OHLC or x.dtype == DataType.AGGREGATED_LIQUIDATIONS:
                    _type_ctor = DataType.OHLC if x.dtype == DataType.OHLC else DataType.AGGREGATED_LIQUIDATIONS
                    if self.synthetic_ohlc_timeframes_types and x.data_timeframe:
                        # - for making life bit easy let's generate all possible frames we can contruct from available base
                        for f in _ext_frames[_ext_frames.searchsorted(pd.Timedelta(x.data_timeframe)) :]:
                            _symbs_lookup[symb][dt := _type_ctor[timedelta_to_str(f)]] = x
                            dtypes.append(dt)
                    else:
                        _symbs_lookup[symb][dt := _type_ctor[x.data_timeframe]] = x
                        dtypes.append(dt)
                else:
                    _symbs_lookup[symb][dt := (x.alias_for_record_type if x.alias_for_record_type else x.dtype)] = x
                    dtypes.append(dt)

            # - for every dtype store symbols and reference to table meta info
            for d in dtypes:
                _dtype_lookup[d] = (set(symbols), x)

        self._symbols_lookup = dict(_symbs_lookup)
        self._dtype_lookup = _dtype_lookup

    def get_data_id(self, dtype: DataType | str = DataType.ALL) -> list[str]:
        d_ids = set()
        for s, di in self._symbols_lookup.items():
            for dt, x in di.items():
                if x.dtype == dtype or dtype == DataType.ALL or str(dtype) == x.alias_for_record_type:
                    d_ids.add(s)
        return list(sorted(d_ids))

    def get_data_types(self, data_id: str) -> list[DataType]:
        return list(self._symbols_lookup.get(data_id, {}).keys())

    def get_time_range(self, data_id: str, dtype: DataType | str) -> tuple[np.datetime64, np.datetime64]:
        (storage_symbols, xtable) = self._dtype_lookup.get(str(dtype).lower(), (set(), None))
        if xtable is None or not storage_symbols:
            raise ValueError(f"Can't find table for {dtype} data !")
        if data_id not in storage_symbols:
            raise ValueError(f"{xtable.table_name} doesn't contain data for {data_id} !")

        _query = f"""(SELECT timestamp FROM "{xtable.table_name}" WHERE symbol='{data_id}' ORDER BY timestamp ASC LIMIT 1)
                        UNION
                   (SELECT timestamp FROM "{xtable.table_name}" WHERE symbol='{data_id}' ORDER BY timestamp DESC LIMIT 1)
                """
        _r = self.pgc.execute(_query)[1]
        _sr = sorted([np.datetime64(_r[0][0]), np.datetime64(_r[1][0])])
        return _sr[0], _sr[1]

    def read(
        self,
        data_id: str | list[str],
        dtype: DataType | str,
        start: str | None,
        stop: str | None,
        chunksize=0,
        **kwargs,
    ) -> Iterator[Transformable] | Transformable:
        # - effectively cut additional info in orderbook (like DataType.ORDERBOOK[0, 10] - just for 10 levels)
        match dtype:
            case DataType.ORDERBOOK:
                x_type = str(DataType.ORDERBOOK).lower()

            case DataType.FUNDAMENTAL:
                x_type = str(DataType.FUNDAMENTAL).lower()

            case _:
                x_type = str(dtype).lower()

        # - get metainfo
        (storage_symbols, xtable) = self._dtype_lookup.get(x_type, (set(), None))
        if xtable is None or not storage_symbols:
            raise ValueError(f"Can't find table for {dtype} data !")

        # - handle symbols
        req_symbols = set(data_id if isinstance(data_id, (list, tuple, set)) else [data_id])
        req_symbols = storage_symbols & req_symbols

        # - check if it has any timeframe for resampling
        _, dt_params = DataType.from_str(dtype)
        resample = dt_params.get("timeframe", "")

        # - handle start / stop
        _start, _stop = handle_start_stop(start, stop)

        # - in case when we want to read by chunks
        if chunksize > 0:

            def _iter_efficient_chunks() -> Iterator[Transformable]:
                time_windows = calculate_time_windows_for_chunking(_start, _stop, resample, chunksize)

                for window_start, window_end in time_windows:
                    yield self._read_data_block(xtable, data_id, req_symbols, dtype, window_start, window_end, resample)

            return _iter_efficient_chunks()

        else:
            return self._read_data_block(xtable, data_id, req_symbols, dtype, _start, _stop, resample)

    def _read_data_block(
        self,
        xtable: xLTableMetaInfo,
        data_id: str | list[str],
        symbols: set[str],
        dtype: DataType | str,
        start: str | pd.Timestamp | None,
        stop: str | pd.Timestamp | None,
        resample: str,
    ) -> Transformable:
        """
        Read data from xtable for given set ot symbols from start to stop
        """
        conditions = []

        if start:
            conditions.append(f"timestamp >= '{start}'")

        if stop:
            conditions.append(f"timestamp < '{stop}'")

        # - get SQL query
        _query = self._prepare_sql_for_dtype(dtype, symbols, xtable, conditions, resample)

        # - process retrieved data
        columns, records = self.pgc.execute(_query)
        data_id_col_idx = find_column_index_in_list(columns, "symbol", "asset")

        # - split received data by symbol
        splitted_records = defaultdict(list)
        for r in records:
            # - keep only requested symbols
            if (symbol := r[data_id_col_idx]) in symbols:
                splitted_records[symbol].append(r)

        # - when requested single symbol just returns single RawData
        if isinstance(data_id, str):
            return RawData(data_id, columns, dtype, splitted_records.get(data_id, []))  # type: ignore

        return RawMultiData([RawData(k, columns, dtype, sr) for k, sr in splitted_records.items()])  # type: ignore

    def _name_in_set(self, name: str, symbols: set[str]) -> str:
        QUOTIFY = lambda ws: map(lambda x: f"'{x.upper()}'", ws)

        # - if requested not too many - just select only them
        if len(symbols) < self.min_symbols_for_all_data_request:
            return f"{name} in ({', '.join(QUOTIFY(symbols))})"

        return ""

    def add_external_builder(self, dtype: str, fn: Callable[[set[str], list[str], str], str]):
        self._external_sql_builders[dtype] = fn

    def _prepare_sql_for_dtype(
        self, dtype: str, symbols: set[str], xtable: xLTableMetaInfo, conditions: list[str], resample: str
    ) -> str:
        match dtype:
            case DataType.OHLC:
                r = """
                    select
                        timestamp,
                        upper(symbol)               as symbol,
                        first(open)                 as open,
                        max(high)                   as high,
                        min(low)                    as low,
                        last(close)                 as close,
                        sum(volume)                 as volume,
                        sum(quote_volume)           as quote_volume,
                        sum(count)                  as count,
                        sum(taker_buy_volume)       as taker_buy_volume,
                        sum(taker_buy_quote_volume) as taker_buy_quote_volume
                    from "{table}" {where} {resample};
                """

                # - select symbols
                conditions.append(self._name_in_set("symbol", symbols))

            case DataType.ORDERBOOK:
                r = """SELECT * FROM "{table}" {where} ORDER BY timestamp ASC"""

                # - process orderbook parameters
                _, params = DataType.from_str(dtype)
                if "depth" in params and params.get("depth") is not None:
                    max_depth = max(int(params["depth"]), 1)
                    conditions.append(f"abs(level) <= {max_depth}")

                # - select assets
                conditions.append(self._name_in_set("symbol", symbols))

            case DataType.FUNDAMENTAL:
                r = """
                    select timestamp, asset, metric, last(value) as value
                    from "{table}" {where} {resample};
                """

                # - check if we want to get specific metrics only
                _, params = DataType.from_str(dtype)
                if "fields" in params and (fields := params.get("fields")) is not None:
                    QUOTIFY = lambda ws: map(lambda x: f"'{x}'", ws)
                    conditions.append(f"metric in ({', '.join(QUOTIFY(fields))})")

                # - select assets
                conditions.append(self._name_in_set("asset", symbols))

            case DataType.FUNDING_PAYMENT:
                r = """
                    SELECT timestamp, symbol, funding_rate, funding_interval_hours
                    FROM "{table}" {where} {resample} ORDER BY timestamp ASC;
                """

                # - select assets
                conditions.append(self._name_in_set("symbol", symbols))

            case DataType.FUNDING_RATE:
                r = """
                    SELECT timestamp, symbol, funding_rate as rate, funding_interval_hours as interval, next_funding_time, mark_price, index_price
                    FROM "{table}" {where} {resample} ORDER BY timestamp ASC;
                """

                # - select assets
                conditions.append(self._name_in_set("symbol", symbols))

            case DataType.LIQUIDATION:
                # - TODO: we need to clarify table's structure
                r = """
                    SELECT timestamp, symbol, quantity, price, side
                    FROM "{table}" {where} ORDER BY timestamp ASC;
                """
                # - select assets
                conditions.append(self._name_in_set("symbol", symbols))

            case DataType.AGGREGATED_LIQUIDATIONS:
                r = (
                    """
                    select
                        {shift} timestamp,
                        upper(symbol)               as symbol,
                        avg(avg_buy_price)          as avg_buy_price,
                        sum(buy_amount)             as buy_amount,
                        sum(buy_count)              as buy_count,
                        sum(buy_notional)           as buy_notional,
                        last(last_buy_price)        as last_buy_price,
                        avg(avg_sell_price)         as avg_sell_price,
                        sum(sell_amount)            as sell_amount,
                        sum(sell_count)             as sell_count,
                        sum(sell_notional)          as sell_notional,
                        last(last_sell_price)       as last_sell_price
                    from "{table}" {where} {resample};
                """
                    if resample
                    else """
                    select
                        timestamp,
                        upper(symbol)               as symbol,
                        avg_buy_price               as avg_buy_price,
                        buy_amount                  as buy_amount,
                        buy_count                   as buy_count,
                        buy_notional                as buy_notional,
                        last_buy_price              as last_buy_price,
                        avg_sell_price              as avg_sell_price,
                        sell_amount                 as sell_amount,
                        sell_count                  as sell_count,
                        sell_notional               as sell_notional,
                        last_sell_price             as last_sell_price
                    from "{table}" {where};
                """
                )

                # - select assets
                conditions.append(self._name_in_set("symbol", symbols))

            case _:
                if dtype in self._external_sql_builders:
                    r = self._external_sql_builders.get(dtype)(symbols, conditions, resample)
                else:
                    r = """SELECT * FROM "{table}" {where};"""

                    # - select assets
                    conditions.append(self._name_in_set("symbol", symbols))

        COMBINE = lambda cs: " and ".join(filter(lambda x: x, cs)) if cs else ""
        where = COMBINE(conditions)
        shift = ""
        if resample:
            n, u = self._convert_time_delta_to_qdb_resample_format(resample)
            # - if request need to be timestamped at right bound of interval when resampling
            shift = f"dateadd('{u}', {n}, timestamp)"
            resample = f"SAMPLE by {n}{u} FILL(NONE)"

        return r.format(
            table=xtable.table_name, where="" if not where else f"where {where}", resample=resample, shift=shift
        )


@storage("qdb")
@storage("questdb")
class QuestDBStorage(IStorage):
    """
    QuestDB storage implementation
    """

    pgc: PGConnectionHelper

    def __init__(
        self,
        host="localhost",
        user="admin",
        password="quest",
        port=8812,
        min_symbols_for_all_data_request: int = 50,
        synthetic_ohlc_timeframes_types: bool = True,
    ) -> None:
        self.pgc = PGConnectionHelper(host, user, password, port)
        self.min_symbols_for_all_data_request = min_symbols_for_all_data_request
        self.synthetic_ohlc_timeframes_types = synthetic_ohlc_timeframes_types

    @property
    def connection_url(self):
        return self.pgc.connection_url

    def __getstate__(self):
        # - close connection before pickling
        state = self.__dict__.copy()
        state["pgc"] = self.pgc.__getstate__()
        return state

    def __setstate__(self, state):
        # - restore state and recreate connection
        self.__dict__.update(state)
        # - recreate PGConnectionHelper with stored connection params
        pgc_state = state["pgc"]
        self.pgc = PGConnectionHelper.__new__(PGConnectionHelper)
        self.pgc.__setstate__(pgc_state)

    def close(self):
        self.pgc.close()

    def __del__(self):
        if hasattr(self, "pgc"):
            self.pgc.__del__()

    def get_exchanges(self) -> list[str]:
        return list(self._read_database_meta_structure().keys())

    def _read_database_meta_structure(self) -> dict[str, dict[str, list[xLTableMetaInfo]]]:
        """
        Read DB structure and build tables mapping:
            exchange -> {market_type -> [xLTableMetaInfo] }
        """
        dbm: dict[str, dict[str, list[xLTableMetaInfo]]] = defaultdict(lambda: defaultdict(list))
        tables = self.pgc.execute("select table_name as name from tables()")[1]
        for t in tables:
            if x := xLTableMetaInfo.decode_table_metadata(t[0]):
                dbm[x.exchange][x.market_type].append(x)

        # - for sanity convert default dict to standard dict
        return {s0: {s1: v1 for s1, v1 in v0.items()} for s0, v0 in dbm.items()}

    def get_market_types(self, exchange: str) -> list[str]:
        exc_i = self._read_database_meta_structure()
        return list(exc_i.get(exchange.upper(), dict()).keys())

    def get_reader(self, exchange: str, market: str) -> IReader:
        e_info = self._read_database_meta_structure()

        if exchange in e_info and market in e_info[exchange]:
            return QuestDBReader(
                exchange,
                market,
                e_info[exchange][market],
                self.pgc,
                synthetic_ohlc_timeframes_types=self.synthetic_ohlc_timeframes_types,
                min_symbols_for_all_data_request=self.min_symbols_for_all_data_request,
                # - for fundamental data we want to search in assets not in symbols
                symbol_column_name="asset" if market.lower() == "fundamental" else "symbol",
            )

        raise ValueError(f"Can't provide data reader for exchange '{exchange}' and market type '{market}'")



"""
Simplified polling adapter to convert CCXT fetch_* methods into watch_* behavior.

This adapter provides a much simpler approach:
- No background tasks or queues
- get_next_data() waits until it's time to poll, then polls synchronously
- Time-aligned polling (e.g., 11:30, 11:35, 11:40 for 5-minute intervals)
- Immediate polling when symbols change
"""

import asyncio
import math
import time
from dataclasses import dataclass
from typing import Any, Callable, Dict, List, Optional, Set

from qubx import logger

# Constants
DEFAULT_POLL_INTERVAL = 300  # 5 minutes
MIN_POLL_INTERVAL = 1  # 1 second minimum
MAX_POLL_INTERVAL = 3600  # 1 hour maximum


@dataclass
class PollingConfig:
    """Configuration for polling adapter."""

    poll_interval_seconds: float = DEFAULT_POLL_INTERVAL

    def __post_init__(self):
        """Validate configuration after initialization."""
        if not MIN_POLL_INTERVAL <= self.poll_interval_seconds <= MAX_POLL_INTERVAL:
            raise ValueError(
                f"poll_interval_seconds must be between {MIN_POLL_INTERVAL} and {MAX_POLL_INTERVAL}, "
                f"got {self.poll_interval_seconds}"
            )


class PollingToWebSocketAdapter:
    """
    Simplified polling adapter that polls synchronously when data is requested.

    Key features:
    - No background tasks or queues
    - Time-aligned polling (respects clock boundaries)
    - Immediate polling when symbols change
    - Thread-safe symbol management
    """

    def __init__(
        self,
        fetch_method: Callable,
        symbols: Optional[List[str]] = None,
        params: Optional[Dict[str, Any]] = None,
        config: Optional[PollingConfig] = None,
    ):
        """
        Initialize the simplified polling adapter.

        Args:
            fetch_method: The CCXT fetch_* method to call
            symbols: Initial list of symbols to watch
            params: Additional parameters for fetch_method
            config: PollingConfig instance (uses default if None)
        """
        self.config = config if config is not None else PollingConfig()
        self.fetch_method = fetch_method
        self.params = params or {}
        self.adapter_id = f"polling_adapter_{id(self)}"

        # Thread-safe symbol management
        self._symbols_lock = asyncio.Lock()
        self._symbols: Set[str] = set(symbols or [])

        # Polling state
        self._last_poll_time: Optional[float] = None
        self._symbols_changed = False  # Flag to trigger immediate poll

        # Statistics
        self._poll_count = 0
        self._error_count = 0

    async def get_next_data(self) -> Dict[str, Any]:
        """
        Get the next available data by waiting until it's time to poll, then polling.

        This method:
        1. Checks if symbols changed (immediate poll)
        2. Calculates when next poll should happen based on time alignment
        3. Waits until that time
        4. Polls and returns fresh data

        Returns:
            Dictionary containing fetched data for symbols
        """
        async with self._symbols_lock:
            current_symbols = list(self._symbols)
            symbols_changed = self._symbols_changed

        # If symbols changed, poll immediately
        if symbols_changed:
            logger.debug(f"Symbols changed, polling immediately for adapter {self.adapter_id}")
            async with self._symbols_lock:
                self._symbols_changed = False
            return await self._poll_now(current_symbols)

        # Calculate wait time for next aligned poll
        wait_time = self._calculate_wait_time()

        if wait_time > 0:
            logger.debug(f"Waiting {wait_time:.1f}s for next poll cycle for adapter {self.adapter_id}")
            await asyncio.sleep(wait_time)

        # Time to poll
        logger.debug(f"Polling now for adapter {self.adapter_id}")
        return await self._poll_now(current_symbols)

    def _calculate_wait_time(self) -> float:
        """
        Calculate how long to wait until the next aligned poll time.

        For intervals >= 1 minute: aligns to clock boundaries (11:30, 11:35, 11:40)
        For intervals < 1 minute: uses simple interval-based timing

        Returns:
            Number of seconds to wait (0 if should poll now)
        """
        current_time = time.time()
        interval_seconds = self.config.poll_interval_seconds

        # First poll is always immediate
        if self._last_poll_time is None:
            return 0

        if interval_seconds >= 60:
            # Time-aligned polling for intervals >= 1 minute using UTC
            # Calculate next boundary based on seconds since epoch
            next_boundary = math.ceil(current_time / interval_seconds) * interval_seconds
            wait_time = next_boundary - current_time
            return max(0, wait_time)
        else:
            # Simple interval-based polling for sub-minute intervals
            next_poll_time = self._last_poll_time + interval_seconds
            wait_time = next_poll_time - current_time
            return max(0, wait_time)

    async def _poll_now(self, symbols: List[str]) -> Dict[str, Any]:
        """
        Perform a poll operation immediately.

        Args:
            symbols: List of symbols to poll for

        Returns:
            Dictionary containing fetched data for symbols
        """
        self._poll_count += 1
        self._last_poll_time = time.time()

        logger.debug(f"Polling {len(symbols) if symbols else 'all'} symbols for adapter {self.adapter_id}")

        try:
            # Filter out adapter-specific parameters
            adapter_params = {"pollInterval", "interval", "updateInterval", "poll_interval_minutes"}
            fetch_params = {k: v for k, v in self.params.items() if k not in adapter_params}

            # Call the fetch method
            result = await self.fetch_method(symbols, **fetch_params)

            logger.debug(f"Poll completed successfully for adapter {self.adapter_id}")
            return result

        except Exception as e:
            self._error_count += 1
            logger.error(f"Poll failed for adapter {self.adapter_id}: {e}")
            raise

    async def update_symbols(self, new_symbols: List[str]) -> None:
        """
        Update the symbol list.

        If symbols changed, the next call to get_next_data() will poll immediately.

        Args:
            new_symbols: New complete list of symbols to watch
        """
        async with self._symbols_lock:
            old_symbols = self._symbols.copy()
            self._symbols = set(new_symbols or [])
            symbols_changed = old_symbols != self._symbols

            if symbols_changed:
                self._symbols_changed = True
                logger.debug(
                    f"Symbols updated for adapter {self.adapter_id}: {len(old_symbols)} -> {len(self._symbols)}"
                )

    async def add_symbols(self, new_symbols: List[str]) -> None:
        """Add symbols to the existing watch list."""
        if not new_symbols:
            return

        async with self._symbols_lock:
            before_count = len(self._symbols)
            self._symbols.update(new_symbols)
            after_count = len(self._symbols)

            if after_count > before_count:
                self._symbols_changed = True
                logger.debug(f"Added {after_count - before_count} symbols to adapter {self.adapter_id}")

    async def remove_symbols(self, symbols_to_remove: List[str]) -> None:
        """Remove symbols from the watch list."""
        if not symbols_to_remove:
            return

        async with self._symbols_lock:
            before_count = len(self._symbols)
            self._symbols.difference_update(symbols_to_remove)
            after_count = len(self._symbols)

            if after_count < before_count:
                self._symbols_changed = True
                logger.debug(f"Removed {before_count - after_count} symbols from adapter {self.adapter_id}")

    def is_watching(self, symbol: Optional[str] = None) -> bool:
        """Check if adapter has symbols configured to watch."""
        if symbol is None:
            return len(self._symbols) > 0
        else:
            return symbol in self._symbols

    def get_statistics(self) -> Dict[str, Any]:
        """Get adapter statistics for monitoring."""
        return {
            "adapter_id": self.adapter_id,
            "symbol_count": len(self._symbols),
            "poll_count": self._poll_count,
            "error_count": self._error_count,
            "last_poll_time": self._last_poll_time,
            "poll_interval_seconds": self.config.poll_interval_seconds,
        }

    async def stop(self) -> None:
        """Stop the adapter (cleanup method for compatibility)."""
        logger.debug(f"Adapter {self.adapter_id} stopped (polled {self._poll_count} times, {self._error_count} errors)")



"""
Redis Streams Exporter for trading data.

This module provides an implementation of ITradeDataExport that exports trading data to Redis Streams.
"""

from concurrent.futures import ThreadPoolExecutor
from typing import Any, Dict, List, Optional, cast

import redis
from redis.typing import EncodableT, FieldT

from qubx import logger
from qubx.core.basics import Instrument, Signal, TargetPosition, dt_64
from qubx.core.interfaces import IAccountViewer, ITradeDataExport
from qubx.exporters.formatters import DefaultFormatter, IExportFormatter


class RedisStreamsExporter(ITradeDataExport):
    """
    Exports trading data to Redis Streams.

    This exporter can be configured to export signals, target positions, and leverage changes.
    """

    def __init__(
        self,
        redis_url: str,
        strategy_name: str,
        export_signals: bool = False,
        export_targets: bool = False,
        export_position_changes: bool = False,
        signals_stream: Optional[str] = None,
        targets_stream: Optional[str] = None,
        position_changes_stream: Optional[str] = None,
        max_stream_length: int = 1000,
        formatter: Optional[IExportFormatter] = None,
        max_workers: int = 2,
        account: Optional[IAccountViewer] = None,
    ):
        """
        Initialize the Redis Streams Exporter.

        Args:
            redis_url: Redis connection URL (e.g., "redis://localhost:6379/0")
            strategy_name: Name of the strategy (used in stream keys if not provided)
            export_signals: Whether to export signals
            export_targets: Whether to export target positions
            export_position_changes: Whether to export position changes
            signals_stream: Custom stream name for signals (default: "strategy:{strategy_name}:signals")
            targets_stream: Custom stream name for target positions (default: "strategy:{strategy_name}:targets")
            position_changes_stream: Custom stream name for position changes (default: "strategy:{strategy_name}:position_changes")
            max_stream_length: Maximum length of each stream
            formatter: Formatter to use for formatting data (default: DefaultFormatter)
            max_workers: Maximum number of worker threads for Redis operations
            account: Optional account viewer to get account information like total capital, leverage, etc.
        """
        self._redis = redis.from_url(redis_url)
        self._strategy_name = strategy_name

        self._export_signals = export_signals
        self._export_targets = export_targets
        self._export_position_changes = export_position_changes

        self._signals_stream = signals_stream or f"strategy:{strategy_name}:signals"
        self._targets_stream = targets_stream or f"strategy:{strategy_name}:targets"
        self._position_changes_stream = position_changes_stream or f"strategy:{strategy_name}:position_changes"

        self._max_stream_length = max_stream_length
        self._formatter = formatter or DefaultFormatter()

        self._instrument_to_previous_leverage = {}

        self._executor = ThreadPoolExecutor(max_workers=max_workers, thread_name_prefix="redis_exporter")

        if account:
            self._instrument_to_previous_leverage = dict(account.get_leverages())

        logger.info(
            f"[RedisStreamsExporter] Initialized for strategy '{strategy_name}' with "
            f"signals: {export_signals}, targets: {export_targets}, position_changes: {export_position_changes}"
        )

    def __del__(self):
        """Clean up resources when the object is destroyed."""
        try:
            self._executor.shutdown(wait=False)
        except:
            pass

    def _prepare_for_redis(self, data: Dict[str, Any]) -> Dict[FieldT, EncodableT]:
        """
        Prepare data for Redis by ensuring all values are strings.

        Args:
            data: Dictionary with string keys and any values

        Returns:
            Dictionary with keys and values compatible with Redis
        """
        # Convert all values to strings and cast the result to the type expected by Redis
        string_dict = {k: str(v) for k, v in data.items()}
        return cast(Dict[FieldT, EncodableT], string_dict)

    def _add_to_redis_stream(self, stream: str, data: Dict[str, Any]) -> None:
        """
        Add data to a Redis stream in a background thread.

        Args:
            stream: The name of the Redis stream
            data: The data to add to the stream
        """
        try:
            # Prepare data for Redis
            redis_data = self._prepare_for_redis(data)

            # Submit the task to the executor
            self._executor.submit(self._add_to_redis_stream_impl, stream, redis_data, self._max_stream_length)
        except Exception as e:
            logger.exception(f"[RedisStreamsExporter] Failed to queue Redis stream operation")

    def _add_to_redis_stream_impl(self, stream: str, redis_data: Dict[FieldT, EncodableT], max_length: int) -> bool:
        """
        Implementation that actually adds data to Redis stream (called from worker thread).

        Args:
            stream: The name of the Redis stream
            redis_data: The data to add to the stream
            max_length: The maximum length of the stream

        Returns:
            bool: True if the operation was successful, False otherwise
        """
        try:
            # Add to Redis stream
            self._redis.xadd(stream, redis_data, maxlen=max_length, approximate=True)
            return True
        except Exception:
            logger.exception(f"[RedisStreamsExporter] Failed to add to Redis stream {stream}")
            return False

    def export_signals(self, time: dt_64, signals: List[Signal], account: IAccountViewer) -> None:
        """
        Export signals to Redis Stream.

        Args:
            time: Timestamp when the signals were generated
            signals: List of signals to export
            account: Account viewer to get account information like total capital, leverage, etc.
        """
        if not self._export_signals or not signals:
            return

        try:
            exported_count = 0
            for signal in signals:
                # Format the signal using the formatter
                data = self._formatter.format_signal(time, signal, account)

                # Skip if formatter returned empty data
                if not data:
                    logger.debug(
                        f"[RedisStreamsExporter] Skipping signal for {signal.instrument} - formatter returned empty data"
                    )
                    continue

                # Add to Redis stream in background thread
                self._add_to_redis_stream(self._signals_stream, data)
                exported_count += 1

            logger.debug(
                f"[RedisStreamsExporter] Queued {exported_count}/{len(signals)} signals for export to {self._signals_stream}"
            )
        except Exception:
            logger.exception("[RedisStreamsExporter] Failed to export signals")

    def export_target_positions(self, time: dt_64, targets: List[TargetPosition], account: IAccountViewer) -> None:
        """
        Export target positions to Redis Stream.

        Args:
            time: Timestamp when the target positions were generated
            targets: List of target positions to export
            account: Account viewer to get account information like total capital, leverage, etc.
        """
        if not self._export_targets or not targets:
            return

        try:
            exported_count = 0
            for target in targets:
                # Format the target position using the formatter
                data = self._formatter.format_target_position(time, target, account)

                # Skip if formatter returned empty data
                if not data:
                    logger.debug(
                        f"[RedisStreamsExporter] Skipping target position for {target.instrument} - formatter returned empty data"
                    )
                    continue

                # Add to Redis stream in background thread
                self._add_to_redis_stream(self._targets_stream, data)
                exported_count += 1

            logger.debug(
                f"[RedisStreamsExporter] Queued {exported_count}/{len(targets)} target positions for export to {self._targets_stream}"
            )
        except Exception:
            logger.exception("[RedisStreamsExporter] Failed to export target positions")

    def export_position_changes(
        self, time: dt_64, instrument: Instrument, price: float, account: IAccountViewer
    ) -> None:
        """
        Export position changes to Redis Stream.

        Args:
            time: Timestamp when the position change occurred
            instrument: The instrument for which the position change occurred
            price: Price at which the position change occurred
            account: Account viewer to get account information like total capital, leverage, etc.
        """
        if not self._export_position_changes:
            return

        previous_leverage = self._instrument_to_previous_leverage.get(instrument, 0.0)
        new_leverage = account.get_leverage(instrument)
        self._instrument_to_previous_leverage[instrument] = new_leverage

        try:
            # Format the leverage change using the formatter
            data = self._formatter.format_position_change(time, instrument, price, account)

            # Add to Redis stream in background thread
            self._add_to_redis_stream(self._position_changes_stream, data)

            logger.debug(
                f"[RedisStreamsExporter] Queued position change for {instrument}: "
                f"{previous_leverage:0.2%} -> {new_leverage:0.2%} @ {price}"
            )
        except Exception:
            logger.exception(f"[RedisStreamsExporter] Failed to export position change")



"""Trade handler for Lighter WebSocket messages"""

from typing import Any

from qubx.core.series import Trade

from .base import BaseHandler


class TradesHandler(BaseHandler[Trade]):
    """
    Handler for Lighter trade messages.

    Converts Lighter trade format to Qubx Trade objects.

    Lighter format:
    ```json
    {
      "channel": "trade:0",
      "type": "update/trade",
      "timestamp": ...,
      "trades": [{
        "trade_id": 212746654,
        "market_id": 1,
        "size": "0.00100",
        "price": "121043.9",
        "is_maker_ask": true,
        "timestamp": 1760041996404,
        ...
      }],
      "liquidation_trades": [...]  // Optional
    }
    ```

    Trade side logic:
    - is_maker_ask=true: maker selling (ask), taker buying -> side=1 (BUY)
    - is_maker_ask=false: maker buying (bid), taker selling -> side=-1 (SELL)
    """

    # Trade side constants (matching Qubx convention)
    SIDE_BUY = 1
    SIDE_SELL = -1

    def __init__(self, market_id: int):
        """
        Initialize trades handler.

        Args:
            market_id: Lighter market ID to handle
        """
        super().__init__()
        self.market_id = market_id

    def can_handle(self, message: dict[str, Any]) -> bool:
        """Check if message is trade for this market"""
        channel = message.get("channel", "")
        msg_type = message.get("type", "")

        # Check if it's a trade message for our market
        expected_channel = f"trade:{self.market_id}"
        is_trade_msg = msg_type in ["subscribed/trade", "update/trade"]

        return channel == expected_channel and is_trade_msg

    def _handle_impl(self, message: dict[str, Any]) -> list[Trade] | None:
        """
        Convert Lighter trade message to Qubx Trade objects.

        Args:
            message: Raw Lighter trade message

        Returns:
            List of Trade objects, or None if no trades in message

        Raises:
            ValueError: If message format is invalid
        """
        trades_list = []

        # Parse regular trades
        regular_trades = message.get("trades")
        if regular_trades:
            for trade_data in regular_trades:
                trade = self._parse_trade(trade_data)
                if trade:
                    trades_list.append(trade)

        # Parse liquidation trades
        liquidation_trades = message.get("liquidation_trades")
        if liquidation_trades:
            for trade_data in liquidation_trades:
                trade = self._parse_trade(trade_data)
                if trade:
                    trades_list.append(trade)

        return trades_list if trades_list else None

    def _parse_trade(self, trade_data: dict[str, Any]) -> Trade | None:
        """
        Parse a single Lighter trade to Qubx Trade.

        Args:
            trade_data: Single trade dict from Lighter

        Returns:
            Trade object, or None if invalid
        """
        try:
            # Extract fields
            trade_id = trade_data.get("trade_id")
            if trade_id is None:
                return None

            timestamp_ms = trade_data.get("timestamp")
            if timestamp_ms is None:
                return None

            price_str = trade_data.get("price")
            size_str = trade_data.get("size")
            is_maker_ask = trade_data.get("is_maker_ask")

            if price_str is None or size_str is None or is_maker_ask is None:
                return None

            # Convert types
            time_ns = int(timestamp_ms * 1_000_000)  # ms -> ns
            price = float(price_str)
            size = float(size_str)

            # Determine side from is_maker_ask
            # is_maker_ask=true: maker was on ask side, taker bought -> BUY
            # is_maker_ask=false: maker was on bid side, taker sold -> SELL
            side = self.SIDE_BUY if is_maker_ask else self.SIDE_SELL

            # Create Trade object
            return Trade(
                time=time_ns,
                price=price,
                size=size,
                side=side,
                trade_id=int(trade_id),
            )

        except (ValueError, TypeError):
            # Log error but don't raise - skip malformed trades
            return None



import asyncio

from .client import LighterClient


class LighterNonceProvider:
    def __init__(self, client: LighterClient):
        self.client = client
        self._nonce = None
        self._lock = asyncio.Lock()

    async def get_nonce(self) -> int:
        async with self._lock:
            if self._nonce is None:
                self._nonce = await self.client.next_nonce()
            else:
                self._nonce += 1
            return self._nonce

    async def resync(self) -> None:
        self._nonce = None



"""
Strategy initialization utilities.

This module provides classes for initializing strategies, including setting up
schedules, warmup periods, and position mismatch resolution.
"""

import uuid
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any, Callable, Dict, Optional

from qubx.core.basics import Instrument, td_64
from qubx.core.interfaces import IStrategyInitializer, ITransferManager, StartTimeFinderProtocol, StateResolverProtocol
from qubx.core.utils import recognize_timeframe

if TYPE_CHECKING:
    from qubx.core.interfaces import IStrategyContext


@dataclass
class BasicStrategyInitializer(IStrategyInitializer):
    """
    Basic implementation of the IStrategyInitializer interface.

    This class stores configuration information set during strategy initialization,
    such as schedules, warmup periods, and position mismatch resolvers.
    """

    # Default values for all fields
    base_subscription: Optional[str] = None
    base_live_subscription: Optional[str] = None
    fit_schedule: Optional[str] = None
    event_schedule: Optional[str] = None
    warmup_period: Optional[str] = None
    start_time_finder: Optional[StartTimeFinderProtocol] = None
    mismatch_resolver: Optional[StateResolverProtocol] = None
    auto_subscribe: Optional[bool] = None
    simulation: Optional[bool] = None
    subscription_warmup: Optional[dict[Any, str]] = None
    data_cache_config: Dict[str, Any] = field(
        default_factory=lambda: {"enabled": True, "prefetch_period": "1w", "cache_size_mb": 1000}
    )

    # Stale data detection configuration
    stale_data_detection_enabled: bool = False
    stale_data_detection_period: Optional[str] = None
    stale_data_check_interval: Optional[str] = None

    # Number of days ahead to check for delisting
    delisting_check_days: int = 1

    # Transfer manager for fund transfers
    _transfer_manager: Optional[ITransferManager] = None

    # Additional configuration that might be needed
    config: Dict[str, Any] = field(default_factory=dict)

    _pending_global_subscriptions: set[str] = field(default_factory=set)
    _pending_instrument_subscriptions: dict[str, set[Instrument]] = field(default_factory=dict)
    _custom_schedules: dict[str, tuple[str, Callable[["IStrategyContext"], None]]] = field(default_factory=dict)

    def set_base_subscription(self, subscription_type: str) -> None:
        self.base_subscription = subscription_type

    def get_base_subscription(self) -> str | None:
        return self.base_subscription

    def set_base_live_subscription(self, subscription_type: str) -> None:
        self.base_live_subscription = subscription_type

    def get_base_live_subscription(self) -> str | None:
        return self.base_live_subscription

    def set_auto_subscribe(self, value: bool) -> None:
        self.auto_subscribe = value

    def get_auto_subscribe(self) -> bool | None:
        return self.auto_subscribe

    def set_fit_schedule(self, schedule: str) -> None:
        self.fit_schedule = schedule

    def get_fit_schedule(self) -> str | None:
        return self.fit_schedule

    def set_event_schedule(self, schedule: str) -> None:
        self.event_schedule = schedule

    def get_event_schedule(self) -> str | None:
        return self.event_schedule

    def set_warmup(self, period: str, start_time_finder: StartTimeFinderProtocol | None = None) -> None:
        self.warmup_period = period
        if start_time_finder is not None:
            self.set_start_time_finder(start_time_finder)

    def get_warmup(self) -> td_64 | None:
        return td_64(recognize_timeframe(self.warmup_period), "ns") if self.warmup_period else None

    def set_start_time_finder(self, finder: StartTimeFinderProtocol) -> None:
        self.start_time_finder = finder

    def get_start_time_finder(self) -> StartTimeFinderProtocol | None:
        return self.start_time_finder

    def get_state_resolver(self) -> StateResolverProtocol | None:
        return self.mismatch_resolver

    def set_state_resolver(self, resolver: StateResolverProtocol) -> None:
        self.mismatch_resolver = resolver

    def set_config(self, key: str, value: Any) -> None:
        self.config[key] = value

    def get_config(self, key: str, default: Any = None) -> Any:
        return self.config.get(key, default)

    @property
    def is_simulation(self) -> bool | None:
        return self.simulation

    def subscribe(self, subscription_type: str, instruments: list[Instrument] | Instrument | None = None) -> None:
        if instruments is None:
            self._pending_global_subscriptions.add(subscription_type)
            return

        if isinstance(instruments, Instrument):
            instruments = [instruments]

        self._pending_instrument_subscriptions[subscription_type].update(instruments)

    def get_pending_global_subscriptions(self) -> set[str]:
        return self._pending_global_subscriptions

    def get_pending_instrument_subscriptions(self) -> dict[str, set[Instrument]]:
        return self._pending_instrument_subscriptions

    def set_subscription_warmup(self, configs: dict[Any, str]) -> None:
        self.subscription_warmup = configs

    def get_subscription_warmup(self) -> dict[Any, str]:
        return self.subscription_warmup if self.subscription_warmup else {}

    def set_data_cache_config(
        self, enabled: bool = True, prefetch_period: str = "1w", cache_size_mb: int = 1000
    ) -> None:
        """
        Configure CachedPrefetchReader for aux data readers.

        Args:
            enabled: Whether to enable data caching
            prefetch_period: Period to prefetch ahead (e.g., "1w", "2d")
            cache_size_mb: Maximum cache size in MB
        """
        self.data_cache_config = {
            "enabled": enabled,
            "prefetch_period": prefetch_period,
            "cache_size_mb": cache_size_mb,
        }

    def get_data_cache_config(self) -> Dict[str, Any]:
        """
        Get CachedPrefetchReader configuration.

        Returns:
            Dictionary with cache configuration
        """
        return self.data_cache_config

    def schedule(self, cron_schedule: str, method: Callable[["IStrategyContext"], None]) -> None:
        """
        Schedule a custom method to be called at specified times.

        Args:
            cron_schedule: Cron-like schedule string (e.g., "0 0 * * *" for daily at midnight)
            method: Method to call - should accept IStrategyContext as parameter
        """
        schedule_id = str(uuid.uuid4())
        self._custom_schedules[schedule_id] = (cron_schedule, method)

    def get_custom_schedules(self) -> dict[str, tuple[str, Callable[["IStrategyContext"], None]]]:
        """
        Get all custom scheduled methods.

        Returns:
            Dictionary mapping schedule IDs to (cron_schedule, method) tuples
        """
        return self._custom_schedules.copy()

    def set_stale_data_detection(self, enabled: bool, detection_period: str = "1d", check_interval: str = "1d") -> None:
        """
        Configure stale data detection settings.

        Args:
            enabled: Whether to enable stale data detection
            detection_period: Period to consider data as stale (e.g., "5Min", "1h"). If None, uses default.
            check_interval: Interval between stale data checks (e.g., "30s", "1Min"). If None, uses default.
        """
        self.stale_data_detection_enabled = enabled
        self.stale_data_detection_period = detection_period
        self.stale_data_check_interval = check_interval

    def get_stale_data_detection_config(self) -> tuple[bool, str | None, str | None]:
        """
        Get current stale data detection configuration.

        Returns:
            tuple: (enabled, detection_period, check_interval)
        """
        return (self.stale_data_detection_enabled, self.stale_data_detection_period, self.stale_data_check_interval)

    def set_delisting_check_days(self, days: int) -> None:
        """
        Set the number of days ahead to check for delisting.

        Args:
            days: Number of days ahead to check for delisting (default: 1)
        """
        self.delisting_check_days = days

    def get_delisting_check_days(self) -> int:
        """
        Get the number of days ahead to check for delisting.

        Returns:
            int: Number of days ahead to check for delisting
        """
        return self.delisting_check_days

    def set_transfer_manager(self, manager: ITransferManager) -> None:
        """
        Set the transfer manager for handling fund transfers between exchanges.

        This is typically used in live mode to inject a transfer service client.
        In simulation mode, a transfer manager is automatically assigned.

        Args:
            manager: Transfer manager implementation
        """
        self._transfer_manager = manager

    def get_transfer_manager(self) -> ITransferManager | None:
        """
        Get the configured transfer manager.

        Returns:
            ITransferManager | None: The transfer manager if set, None otherwise
        """
        return self._transfer_manager
